import pytest
import asyncio
from typing import *

from base_aux.base_monkeys.m1_ga_self import Monkey_GaSelf_CallResult


# =====================================================================================================================
def test_monkey__value_and_eq():
    victim = Monkey_GaSelf_CallResult(5)

    print()
    print("-"*50)
    print(f"{victim.count_columns()=}")
    print("-"*50)

    assert victim != 5
    assert victim == Monkey_GaSelf_CallResult(5)
    assert victim() == 5
    assert victim() != Monkey_GaSelf_CallResult(5)

    assert victim.hello == victim
    assert victim.hello != 5
    assert victim.hello() == 5
    assert victim.hello() != 100
    assert victim.hello() != victim
    assert victim.hello() != Monkey_GaSelf_CallResult(5)

    assert victim.hello.world == victim
    assert victim.hello.world() == 5
    assert victim.hello.world() != 100

    assert victim.hello.world.count_columns() == 5
    assert victim.hello.world.count_columns() != 100


# =====================================================================================================================



import pytest
import asyncio
from typing import *

from base_aux.base_nest_dunders.m4_ga_self import NestGa_Self


# =====================================================================================================================
def test_1_clear():
    victim = NestGa_Self()
    assert victim == victim.hello
    assert victim == victim.hello.world


def test_2_nested():

    class Victim(NestGa_Self):
        attr = 1
        def meth(self, param=None):
            return param

    victim = Victim()
    assert victim.attr == victim.hello.attr
    assert victim.attr == victim.hello.world.attr

    assert victim.meth(5) == victim.hello.meth(5)
    assert victim.meth(5) == victim.hello.world.meth(5)


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m4_gsai_annots import NestGAI_AnnotAttrIC
from base_aux.aux_attr.m4_kits import *


# =====================================================================================================================
class Victim1:
    ATTR1: int
    ATTR2: int = 2
    ATTR01 = 11


class Victim2(Victim1):
    ATTR3: int
    ATTR4: int = 4
    ATTR02 = 22


VICTIM1_DICT_TYPES = {"ATTR1": int, "ATTR2": int}
VICTIM2_DICT_TYPES = {"ATTR1": int, "ATTR2": int, "ATTR3": int, "ATTR4": int}

VICTIM1_DICT_VALUES = {"ATTR2": 2}
VICTIM2_DICT_VALUES = {"ATTR2": 2, "ATTR4": 4}

victim1 = Victim1()
victim2 = Victim2()


# =====================================================================================================================
class VictimDirect_Ok(NestGAI_AnnotAttrIC):
    ATTR1: int = 1
    ATTR2: int = 2


class VictimDirect_Fail(NestGAI_AnnotAttrIC):
    ATTR1: int
    ATTR2: int = 2

# -----------------------------------------------
class VictimNested_FailParent(VictimDirect_Fail):
    ATTR2: int = 2

class VictimNested_FailChild(VictimDirect_Ok):
    ATTR3: int

# -----------------------------------------------
class VictimNested_OkParent(VictimDirect_Ok):
    ATTR1: int

class VictimNested_OkCHild(VictimDirect_Fail):
    ATTR1: int = 1

# -----------------------------------------------
class DictDirect_Ok(dict, NestGAI_AnnotAttrIC):
    ATTR1: int = 1
    ATTR2: int = 2

class DictDirect_Fail(dict, NestGAI_AnnotAttrIC):
    ATTR1: int


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        (Victim1(), ["ATTR1", ]),
        (Victim2(), ["ATTR1", "ATTR3"]),

        (VictimDirect_Ok(), []),
        (VictimDirect_Fail(), ["ATTR1", ]),

        (VictimNested_FailParent(), ["ATTR1", ]),
        (VictimNested_FailChild(), ["ATTR3", ]),

        (VictimNested_OkParent(), []),
        (VictimNested_OkCHild(), []),

        (DictDirect_Ok(), []),
        (DictDirect_Fail(), ["ATTR1", ]),
    ]
)
def test__annot__get_not_defined(source, _EXPECTED):
    func_link = AttrAux_AnnotsAll(source).annots__get_not_defined
    Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        (VictimDirect_Ok(), True),
        (VictimDirect_Fail(), False),

        (VictimNested_FailParent(), False),
        (VictimNested_FailChild(), False),

        (VictimNested_OkParent(), True),
        (VictimNested_OkCHild(), True),

        (DictDirect_Ok(), True),
        (DictDirect_Fail(), False),
    ]
)
def test__annot__check_all_defined(source, _EXPECTED):
    func_link = AttrAux_AnnotsAll(source).annots__check_all_defined
    Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        (VictimDirect_Ok(), True),
        (VictimDirect_Fail(), Exception),

        (VictimNested_FailParent(), Exception),
        (VictimNested_FailChild(), Exception),

        (VictimNested_OkParent(), True),
        (VictimNested_OkCHild(), True),

        (DictDirect_Ok(), True),
        (DictDirect_Fail(), Exception),
    ]
)
def test__annot__raise_if_not_defined(source, _EXPECTED):
    func_link = AttrAux_AnnotsAll(source).annots__check_all_defined_or_raise
    Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================
# =====================================================================================================================
# =====================================================================================================================
# =====================================================================================================================
# =====================================================================================================================
# =====================================================================================================================
class Test__1:
    # @classmethod
    # def setup_class(cls):
    #     pass
    #     cls.victim1 = Victim1()
    #     cls.victim2 = Victim2()    # @classmethod

    # =================================================================================================================
    # def test__anycase_getattr(self):
    #     assert self.victim2.ATTR2 == 2
        # assert self.victim2.attr2 == 2

    # def test__anycase_getitem(self):
    #     assert self.victim2["ATTR2"] == 2
        # assert self.victim2["attr2"] == 2

    # =================================================================================================================
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (victim1, VICTIM1_DICT_TYPES),
            (victim2, VICTIM2_DICT_TYPES),
        ]
    )
    def test__dict_types(self, source, _EXPECTED):
        func_link = AttrAux_AnnotsAll(source).dump_dict__annot_types
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # =================================================================================================================
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (victim1, VICTIM1_DICT_VALUES),
            (victim2, VICTIM2_DICT_VALUES),
        ]
    )
    def test__dict_values(self, source, _EXPECTED):
        func_link = AttrAux_AnnotsAll(source).dump_dict
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # =================================================================================================================
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (victim1, list(VICTIM1_DICT_VALUES.values())),
            (victim2, list(VICTIM2_DICT_VALUES.values())),
        ]
    )
    def test__iter_values(self, source, _EXPECTED):
        func_link = list(AttrAux_AnnotsAll(source).iter__annot_values())
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # =================================================================================================================
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (victim1, False),
            (victim2, False),
        ]
    )
    def test__all_defined(self, source, _EXPECTED):
        func_link = AttrAux_AnnotsAll(source).annots__check_all_defined
        Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================
def test__all_defined2():
    class Victim11(NestGAI_AnnotAttrIC):
        ATTR1: int = 1
        ATTR2: int = 2
        ATTR01 = 11

    victim11 = Victim11()
    assert AttrAux_AnnotsAll(victim1).annots__check_all_defined() == False
    assert AttrAux_AnnotsAll(victim11).annots__check_all_defined() == True


# =====================================================================================================================
class VictimSet:
    A0 = 1
    A1: int
    A2: int = 1


@pytest.mark.parametrize(
    argnames="data, _EXPECTED",
    argvalues=[
        # only_annot=FALSE
        (dict(), (1, Exception, 1, Exception)),
        (dict(A0=11), (11, Exception, 1, Exception)),
        (dict(a0=11), (11, Exception, 1, Exception)),

        # (dict(a0=11, a1=11, a2=11, a3=11), (11, 11, 11, Exception)),
        # (dict(a0=11, a1=11, a2=11, a3=11), (11, 11, 11, 11)),
        (dict(A0=11, A1=11, A2=11, A3=11), (11, 11, 11, 11)),
    ]
)
def test__set(data, _EXPECTED):
    victim = VictimSet()

    AttrAux_AnnotsAll(victim).sai__by_kwargs(**data)
    for index in range(4):
        Lambda(getattr, victim, f"A{index}").check_expected__assert(_EXPECTED[index])


# =====================================================================================================================
def test__annots_ensure():
    victim = AttrDumped()
    try:
        victim.__annotations__
    except:
        pass
    else:
        assert False

    try:
        victim.__annotations__
    except:
        pass
    else:
        assert False

    AttrAux_AnnotsAll(victim).annots__ensure()
    assert victim.__annotations__ == {}


def test__annots_append():
    # ---------------------------------------------------------
    victim = AttrAux_AnnotsAll().annots__append_with_values()
    assert "astr" not in victim.__annotations__
    assert "aint" not in victim.__annotations__
    assert victim.__annotations__ == dict()

    victim = AttrAux_AnnotsAll(victim).annots__append_with_values(astr=str, aint=1)
    assert "astr" in victim.__annotations__
    assert "aint" in victim.__annotations__
    assert victim.__annotations__ == dict(astr=str, aint=int)

    try:
        victim.astr
    except:
        pass
    else:
        assert False

    assert victim.aint == 1

    # ---------------------------------------------------------
    victim2 = AttrAux_AnnotsAll().annots__append_with_values(astr="hello")
    try:
        victim2.aint
    except:
        pass
    else:
        assert False

    assert victim2.astr == "hello"
    assert victim.__annotations__ != victim2.__annotations__

    # ---------------------------------------------------------
    victim3 = AttrAux_AnnotsAll(victim).annots__append_with_values(astr="hello")
    assert victim3.aint == 1
    assert victim3.astr == "hello"

    assert victim.__annotations__ == victim3.__annotations__


# =====================================================================================================================
class Test__SpecialObjects:
    def test__NamedTuple(self):
        class Victim(NamedTuple):
            ATTR1: int
            ATTR2: int = 2

        try:
            victimNT = Victim() # will not need! raised just on NamedTuple!
        except:
            pass
        else:
            assert False

        victimNT = Victim(1)

        assert AttrAux_AnnotsAll(victimNT).annots__get_not_defined() == []
        assert AttrAux_AnnotsAll(victimNT).annots__check_all_defined() == True
        assert AttrAux_AnnotsAll(victimNT).dump_dict__annot_types() == {"ATTR1": int, "ATTR2": int, }
        # assert AttrAux_AnnotsAll(victimNT).dump_dict() == {"ATTR1": 1, "ATTR2": 2, }

    @pytest.mark.skip
    def test__DataClass(self):
        pass


# =====================================================================================================================
def test__zero():
    pass


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from typing import *
import pathlib

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from base_aux.pyqt.m0_static import COLOR_TUPLE_RGB


# =====================================================================================================================
# TODO: add COMBO with context-manu for add+delete+lastValue
# TODO: create blank Objects to change state!!! GREEn/RED/Yellow/bare


# ICONS ===============================================================================================================
class Icons:
    """
    this is just an example for keeping set of icons!
    """
    _path: pathlib.Path = pathlib.Path().parent.joinpath('images')

    def __init__(self):
        # JPG
        self.logo_square_200: QIcon = QIcon(str(self._path.joinpath('logo_square_200.jpg')))

        # PNG
        self.load_settings: QIcon = QIcon(str(self._path.joinpath('load_settings.png')))
        self.init: QIcon = QIcon(str(self._path.joinpath('init.png')))
        self.ok: QIcon = QIcon(str(self._path.joinpath('ok.png')))
        self.error: QIcon = QIcon(str(self._path.joinpath('error.png')))
        self.save: QIcon = QIcon(str(self._path.joinpath('save.png')))

        # SVG
        self.arrow_right: QIcon = QIcon(str(self._path.joinpath('arrow_right.svg')))
        self.folder2_open: QIcon = QIcon(str(self._path.joinpath('folder2_open.svg')))
        self.gear: QIcon = QIcon(str(self._path.joinpath('gear.svg')))


# COLOR ===============================================================================================================
class WgtColorChange(QWidget):
    def _color_set_rgb(self, color=None):
        if color:
            self.setStyleSheet(f"background: rgb{color}")
        else:
            self.setStyleSheet(f"")

    def color_clear(self):
        self._color_set_rgb(None)

    def color_set__light_green(self):
        self._color_set_rgb(COLOR_TUPLE_RGB.LIGHT_GREEN)

    def color_set__light_red(self):
        self._color_set_rgb(COLOR_TUPLE_RGB.LIGHT_RED)

    def color_set__light_yellow(self):
        self._color_set_rgb(COLOR_TUPLE_RGB.LIGHT_YELLOW)


class WgtColored:
    BTN: QPushButton = type("BTN", (QPushButton, WgtColorChange), {})
    LE: QPushButton = type("LE", (QLineEdit, WgtColorChange), {})
    CB: QPushButton = type("CHB", (QComboBox, WgtColorChange), {})
    LBL: QPushButton = type("LBL", (QLabel, WgtColorChange), {})


# =====================================================================================================================
class QPushButton_Checkable(QPushButton):
    def __init__(self, *args, **kwargs):
        self.__text_toggle: list[str] = ["PushDown", "Release"]  # dont move to class method!!! will not work correct!

        # make available init with list!!! -------------------
        args = list(args)
        for index, arg in enumerate(args):
            if isinstance(arg, str):
                self.__text_toggle[0] = self.__text_toggle[1] = arg
                break
            elif isinstance(arg, list) and len(arg) == 2:
                self.__text_toggle = arg
                args[index] = arg[0]
                break
        super().__init__(*args, **kwargs)

        self.__text_toggle_state()

        # work --------------------------------
        self.setCheckable(True)
        self.toggled.connect(self.__text_toggle_state)

    def __text_toggle_state(self, state=None):
        if state is None:
            state = int(self.isChecked())

        self.setText(self.__text_toggle[state])

    def setText(self, text):
        index = int(self.isChecked())
        self.__text_toggle[index] = text
        super().setText(text)


# TESTS ===============================================================================================================
class __TestExample(QWidget):
    def __init__(self):
        super().__init__()

        self.btn1 = QPushButton_Checkable("1")
        self.btn2 = QPushButton_Checkable(["21", "22"])
        self.btn3 = QPushButton_Checkable()

        layout = QVBoxLayout()
        layout.addWidget(self.btn1)
        layout.addWidget(self.btn2)
        layout.addWidget(self.btn3)
        self.setLayout(layout)
        self.show()

    def print_msg(self):
        print()
        print(self.btn1.isChecked())
        print(self.btn2.isChecked())
        print(self.btn3.isChecked())


# =====================================================================================================================
if __name__ == "__main__":
    app = QApplication([])
    wgt = __TestExample()
    print(app.exec_())


# =====================================================================================================================



from typing import *
import time
# import datetime

from PyQt5.QtCore import QThread, pyqtSignal

from base_aux.servers.m1_client_requests import *
from base_aux.loggers.m2_logger import *
from base_aux.path2_file.m3_filetext import *
from base_aux.aux_datetime.m2_datetime import *

# ---------------------------------------------------------------------------------------------------------------------
from base_aux.testplans.tc import Base_TestCase
from base_aux.testplans.devices_base import Base_Device
from base_aux.testplans.devices_kit import DeviceKit
from base_aux.testplans.gui import Base_TpGui
from base_aux.testplans.api import TpApi_FastApi
from base_aux.testplans.stand import Base_Stand


# =====================================================================================================================
class TpManager(Logger, QThread):
    signal__tp_start = pyqtSignal()
    signal__tp_stop = pyqtSignal()
    signal__tp_finished = pyqtSignal()
    signal__devs_detected = pyqtSignal()

    _signal__tp_reset_duts_sn = pyqtSignal()

    # SETTINGS ------------------------------------------------------
    TP_RUN_INFINIT: bool | None = None     # True - when run() started - dont stop!
    TP_RUN_INFINIT__TIMEOUT: int = 3

    _TC_RUN_SINGLE: bool | None = None

    START__GUI_AND_API: bool = True

    API_SERVER__START: bool = True
    API_SERVER__CLS: type[TpApi_FastApi] = TpApi_FastApi
    api_server: TpApi_FastApi

    GUI__START: bool = True
    GUI__CLS: type[Base_TpGui] = Base_TpGui

    api_client: Client_RequestsStack = Client_RequestsStack()   # todo: USE CLS!!! + add start

    # AUX -----------------------------------------------------------
    STANDS: 'Stands'
    STAND: Base_Stand

    __tc_active: Optional[type[Base_TestCase]] = None

    # =================================================================================================================
    @property
    def tc_active(self) -> type[Base_TestCase] | None:
        return self.__tc_active

    @tc_active.setter
    def tc_active(self, value: type[Base_TestCase] | None) -> None:
        self.__tc_active = value

    def tp__check_active(self) -> bool:
        result = self.tc_active is not None and self.tc_active.STATE_ACTIVE__CLS == EnumAdj_ProcessStateActive.STARTED
        return result

    # =================================================================================================================
    def __init__(self):
        super().__init__()

        self.slots_connect()
        self.init_post()

        # FINAL FREEZE ----------------
        if self.START__GUI_AND_API:
            self.start__gui_and_api()

    def init_post(self) -> None | NoReturn:     # DO NOT DELETE
        """
        GOAL
        ----
        additional user init method

        SPECIALLY CREATED FOR
        ---------------------
        serial devises resolve addresses
        """

    def start__gui_and_api(self) -> None:
        if self.API_SERVER__START:
            self.LOGGER.debug("starting api server")
            self.api_server = self.API_SERVER__CLS(data=self)
            self.api_server.start()

        # last execution --------------------------------------
        if self.GUI__START:
            self.LOGGER.debug("starting gui")
            self.gui = self.GUI__CLS(self)

            # this will BLOCK process
            # this will BLOCK process
            # this will BLOCK process
            # this will BLOCK process
            # this will BLOCK process
            self.gui.run()
        elif self.API_SERVER__START:
            self.api_server.wait()  # it is ok!!!

    def slots_connect(self) -> None:
        self.signal__tp_start.connect(self.start)
        self.signal__tp_stop.connect(self.terminate)

        Base_TestCase.signals.signal__tc_state_changed.connect(self.post__tc_results)

    # =================================================================================================================
    @classmethod
    def stand__init(cls, item: Base_Stand = None) -> None:
        if item is not None:
            cls.STAND = item

    def tcs_clear(self) -> None:
        self.STAND.TIMESTAMP_START = DateTimeAux()
        for tc_cls in self.STAND.TCSc_LINE:
            tc_cls.clear__cls()

    # =================================================================================================================
    def tp__startup(self) -> bool:
        """
        Overwrite with super! super first!
        """
        self.STAND.DEV_LINES("connect__only_if_address_resolved")  #, group="DUT")   # dont connect all here! only in exact TC!!!!????
        return True

    def tp__teardown(self) -> None:
        """
        Overwrite with super! super last!
        """
        if self.tc_active and (self.tc_active.STATE_ACTIVE__CLS == EnumAdj_ProcessStateActive.STARTED or self.tc_active.result__teardown_cls is None):
            self.tc_active.terminate__cls()
        if not self._TC_RUN_SINGLE:
            self.tc_active = None

        self.STAND.TIMESTAMP_STOP = DateTimeAux()
        self.STAND.DEV_LINES("disconnect")

        # self.signal__tp_finished.emit()   # dont place here!!!

    # =================================================================================================================
    def terminate(self) -> None:
        pass

        need_msg: bool = False
        if self.isRunning():
            need_msg = True
            super().terminate()

        # TERMINATE CHILDS!!! ---------------------
        # ObjectInfo(self.currentThread()).print()    # cant find childs!!!

        # finish active ----------------------------
        if self.tc_active:
            self.tc_active.terminate__cls()

        # finish ----------------------------
        self.tp__teardown()
        if need_msg:
            self.signal__tp_finished.emit()

    def run(self) -> None:
        self.LOGGER.debug("TP START")
        if self.tp__check_active():
            return

        cycle_count = 0
        while True:
            if not self._TC_RUN_SINGLE:
                self.tcs_clear()

            cycle_count += 1

            if self.tp__startup():
                # tcs_to_execute = list(filter(lambda x: not x.SKIP, self.STAND.TCSc_LINE))
                tcs_to_execute = self.STAND.TCSc_LINE

                if self._TC_RUN_SINGLE:
                    if not self.tc_active:
                        try:
                            self.tc_active = list(filter(lambda x: not x.SKIP, self.STAND.TCSc_LINE))[0]
                        except:
                            self.tc_active = self.STAND.TCSc_LINE[0]

                    self.tc_active.run__cls()

                else:
                    # MULTY
                    for tc_new in tcs_to_execute:     # TODO: place cls_prev into TcBaseCls!!! and clear on finish???
                        if tc_new.SKIP:
                            continue

                        # SWITCH/ROLL ---------
                        tc_prev = self.tc_active
                        self.tc_active = tc_new

                        tc_executed__result = self.tc_active.run__cls(cls_prev=tc_prev)

                        if tc_executed__result is False:
                            break

            # EXIT/STOP LAST TC
            # if self.tc_active and self.tc_active.STATE_ACTIVE__CLS != None:
            #     self.tc_active.teardown__cls()
            # FINISH TP CYCLE ---------------------------------------------------
            self.tp__teardown()
            self.LOGGER.debug("TP FINISH")

            # RESTART -----------------------------------------------------
            if not self.TP_RUN_INFINIT:
                break

            time.sleep(self.TP_RUN_INFINIT__TIMEOUT)

        # FINISH TP TOTAL ---------------------------------------------------
        self.signal__tp_finished.emit()

    # =================================================================================================================
    # FIXME: REF!!!
    def post__tc_results(self, tc_inst: Base_TestCase) -> None:
        # CHECK ------------------------------------------
        if not self.api_client or tc_inst.result is None:
            return

        # WORK ------------------------------------------
        try:
            tc_results = tc_inst.tci__get_result()
        except:
            tc_results = {}

        body = {
            **self.STAND.stand__get_info__general(),
            **tc_results,
        }
        self.api_client.send(body=body)


# =====================================================================================================================
class TpInsideApi_Runner(TpApi_FastApi):
    """
    REASON:
    in windows Base_TestCase works fine by any variance GUI__START/API_SERVER__START
    in Linux it is not good maybe cause of nesting theme=Thread+Async+Threads

    so this is the attempt to execute correctly TP in Linux by deactivating GUI and using theme=Async+Threads

    UNFORTUNATELY: ITS NOT WORKING WAY for linux!!!
    """
    TP_CLS: type[TpManager] = TpManager

    def __init__(self, *args, **kwargs):

        self.TP_CLS.START__GUI_AND_API = False
        self.data = self.TP_CLS()

        super().__init__(*args, **kwargs)
        self.run()


# =====================================================================================================================



"""
TODO: try to use simple TableObj instead!!!

IDEAS
-----
TableLine
    - work with all insts in one group
    - unnamed collection
TableKit
    - work with all insts in all groups
    - named collection (dict-like)
TableColumn
    - work with all insts in one column (useful for instances like TestCase)

USAGE
-----
1. creating Table
2. using like simple collections (iterable/gi)
"""

# =====================================================================================================================
from base_aux.base_values.m3_exceptions import *
from base_aux.aux_argskwargs.m1_argskwargs import *


# =====================================================================================================================
class TableLine:
    """
    GOAL
    ----
    smth like a group with several or one instances

    GI-access to elements.
        RETURN
            if INSTS multy - return source[index]
            otherwise - INSTS[0]

    SPECIALLY CREATED FOR
    ---------------------
    simplifying work with Breeder like object!
    (most important difference is working with already generated Elements!)
    """
    INSTS: tuple[Any, ...]

    def __init__(self, *insts: Any) -> None:
        """
        if one instance for all Columns - use one instance
        if used several instances - use exact count - each inst for each Columns
        """
        self.INSTS = insts

    def __iter__(self) -> Iterable[Any]:
        """
        GOAL
        ----
        iter all instances in line

        NOTE-IMPORTANT
        --------------
        if instance IS (not EQ!) previous instance - skip!
        it NEED for applying several ATC for some groups (3 ATC for 12 PTB, so there are 1 ATC for 4 PTB)

        SO as result
            self.INSTS != [*self]
            len(self.INSTS) == len(self) >= [*self]
        """
        inst_prev = None
        for inst in self.INSTS:
            if inst is not inst_prev:
                inst_prev = inst
                yield inst

    def __contains__(self, item) -> bool:
        return item in self.INSTS

    def __getitem__(self, index: int) -> Any | NoReturn:
        """
        GOAL
        ----
        access to exact instance by index
        """
        if len(self.INSTS) == 1:
            return self.INSTS[0]
        else:
            return self.INSTS[index]

    # def __getattr__(self, item: str) -> Self | NoReturn:    # NOTE: DONT USE IT!!! CANT compose result with called value
    #     """
    #     GOAL
    #     ----
    #     used as calling methods on all INSTS
    #     """
    #     result = []
    #     for inst in self.INSTS:
    #         try:
    #             result_i = getattr(inst, item)
    #         except Exception as exc:
    #             result_i = exc
    #
    #         result.append(result_i)
    #
    #     return TableLine(*result)

    def __len__(self) -> int:
        """
        GOAL
        ----
        return number of line instances

        if one instance for all Columns - return 1
        """
        return len(self.INSTS)

    def __call__(self, meth: str, *args, **kwargs) -> list[Any | Exception]:        # TODO: APPLY TableLine as result???
        """
        GOAL
        ----
        call method on all instances
        """
        results = []
        inst_prev = None
        for inst in self.INSTS:
            if inst is not inst_prev:
                inst_prev = inst
                try:
                    inst_meth = getattr(inst, meth)
                    ints_result = inst_meth(*args, **kwargs)
                except Exception as exc:
                    ints_result = exc
            else:
                ints_result = results[-1]

            results.append(ints_result)

        return results

    @property
    def COUNT(self) -> int:
        """
        preferred using direct LEN???
        """
        return len(self.INSTS)

    def __eq__(self, other: Any | Self) -> bool:    # DECIDE: delete? it is not needed???
        """
        CREATED SPECIALLY FOR
        ---------------------
        just testing perpose! not a real tip!
        """
        if isinstance(other, TableLine):
            if self.COUNT == other.COUNT:
                for inst1, inst2 in zip(self.INSTS, other.INSTS):
                    if inst1 != inst2:
                        return False
                return True
            else:
                return False
        else:
            # VAR-1=BAD! not clear and DIFFICALT!
            # what if (INSTS-Single and Other-Multy) or ViceVersa
            # return other == self.INSTS

            # VAR-2=best way
            return False


# =====================================================================================================================
class TableKit:     # todo: add AttrsKit nesting???
    """
    GOAL
    ----
    just as object keeping sets for all lines

    USAGE
    =====
    two ways to define object
    -------------------------
        1=by direct set cls attrs
        2=by init kwargs

    create/use Instance
    -------------------
    """
    _count_columns: int = 1

    def __init__(self, **lines: TableLine) -> None | NoReturn:
        self._init_new_lines(**lines)
        self._init_count_columns()
        self._check_same_counts()

    # -----------------------------------------------------------------------------------------------------------------
    def _init_new_lines(self, **lines: TableLine) -> None | NoReturn:
        # TODO: add/extend in annotations???? - not really need!

        for name, value in lines.items():
            if isinstance(value, TableLine):
                setattr(self, name, value)
            else:
                msg = f"{value=} is not TableLine type"
                raise Exc__WrongUsage(msg)

    def _init_count_columns(self) -> None:
        for name, line in self.items():
            self.COUNT_COLUMNS = line.COUNT

    def _check_same_counts(self) -> None | NoReturn:
        for name, line in self.items():
            if line.COUNT not in [self.COUNT_COLUMNS, 1]:
                msg = f"{name=}/{line.COUNT=}/{self.COUNT_COLUMNS=}"
                raise Exc__WrongUsage(msg)

    # -----------------------------------------------------------------------------------------------------------------
    def __len__(self) -> int:
        """
        GOAL
        ----
        return count
        """
        return len(self.names())

    @property
    def COUNT_COLUMNS(self) -> int:
        return self._count_columns

    @COUNT_COLUMNS.setter
    def COUNT_COLUMNS(self, new: int) -> None | NoReturn:
        if new == 1:
            return

        if self._count_columns == 1:
            self._count_columns = new
        elif self._count_columns != new:
            msg = f"{new=}/{self.COUNT_COLUMNS=}"
            raise Exc__WrongUsage(msg)

    def size(self) -> tuple[int, int]:
        return len(self), self.COUNT_COLUMNS

    # -----------------------------------------------------------------------------------------------------------------
    def __contains__(self, item: str) -> bool:
        """
        GOAL
        ----
        check just name line exist in lines
        """
        return item in self.names()

    def __getitem__(self, item: str) -> TableLine | NoReturn:
        """
        GOAL
        ----
        access to LINE over str name!
        just as additional ability! - but its is really need in GUI tableModel
        """
        result = getattr(self, item)
        if isinstance(result, TableLine):
            return result
        else:
            msg = f"no TableLine item in LINES [{item=}]"
            raise Exc__Addressing(msg)

    # -----------------------------------------------------------------------------------------------------------------
    def items(self) -> Iterable[tuple[str, TableLine]]:
        """
        NOTE/CAREFUL!
        ----
        iterate in DIR ORDER!!! not as defined!
        """
        # TODO: apply Annotated aatrs only???   - not really need!
        for name in dir(self):
            # print(f"items={name=}")
            if name.startswith("_"):
                continue
            value = getattr(self, name)
            if isinstance(value, TableLine):
                yield name, value

    def names(self) -> list[str]:   # DONT USE SET!!!
        result = []
        for name, value in self.items():
            result.append(name)
        return result

    def values(self) -> list[TableLine]:
        """
        NOTE
        ----
        lineInstances (TableLine())! not internal Line instances(TableLine().INSTS)!
        """
        result = []
        for name, value in self.items():
            result.append(value)
        return result

    def iter_lines_insts(self) -> Iterable[Any]:
        """
        GOAL
        ----
        iter ALL instances from all LINES!
        """
        for line in self.values():
            yield from line

    # -----------------------------------------------------------------------------------------------------------------
    def __call__(self, meth: str, *args, **kwargs) -> dict[str, list[Any | Exception]]:
        """
        GOAL
        ----
        call method on all lines
        """
        results = {}
        for name, line in self.items():
            results.update({name: line(meth, *args, **kwargs)})

        return results


# =====================================================================================================================
class TableColumn:
    """
    GOAL
    ----
    replace/ref breederObject!
    access to exact instance in line by simple name (implying index)
    """
    LINES: TableKit = TableKit()   # access for all lines!
    INDEX: int

    def __init__(self, index: int, lines: TableKit = None) -> None | NoReturn:
        if lines is not None:
            self.LINES = lines

        if not isinstance(self.LINES, TableKit):
            msg = f"{self.LINES=} is non type(TableKit)"
            raise Exc__WrongUsage(msg)

        if index + 1 > self.LINES.COUNT_COLUMNS:
            msg = f"{index=}/{self.LINES.COUNT_COLUMNS=}"
            raise Exc__Addressing(msg)

        self.INDEX = index

    def __getattr__(self, item: str) -> Any | NoReturn:
        """
        GOAL
        ----
        get index from exact line by name

        """
        # if not hasattr(self, "LINES"):
        #     raise Exception("hello")

        line: TableLine = getattr(self.LINES, item)
        if isinstance(line, TableLine):     # as Line
            return line[self.INDEX]

        elif isinstance(line, list):        # as list                       # todo: decide is it really need?
            return line[self.INDEX]

        else:                               # as direct ATTR from LINES     # todo: decide is it really need? couse all LineAttrs would return
            return line


# =====================================================================================================================



from base_aux.base_nest_dunders.m6_eq2_cls import *
from base_aux.aux_attr.m4_kits import *


# =====================================================================================================================
def test__incorrect_zero():
    class Victim0:
        ATTR = 1

    class VictimEq(Nest_EqCls):
        ATTR = 1

    assert Victim0 != VictimEq
    assert Victim0 != VictimEq()
    assert Victim0() != VictimEq()
    assert Victim0() != VictimEq

    assert VictimEq != Victim0
    assert VictimEq != Victim0()
    assert VictimEq() != Victim0()
    assert VictimEq() != Victim0


def test__incorrect_direct():
    class Victim0:
        ATTR = 1

    class VictimEq(Nest_EqCls):
        ATTR = 1

        @classmethod
        @property
        def _EQ_CLS__VALUE(cls) -> Any:
            return cls.ATTR

    assert Victim0 != VictimEq
    assert Victim0 != VictimEq()
    assert Victim0() != VictimEq()
    assert Victim0() != VictimEq

    assert VictimEq != Victim0
    assert VictimEq != Victim0()
    assert VictimEq() != Victim0()
    assert VictimEq() != Victim0


# =====================================================================================================================
def test__correct_false():
    class Victim0:
        ATTR = 1

    class VictimEq(Nest_EqCls):
        ATTR = 1

        @classmethod
        @property
        def _EQ_CLS__VALUE(cls) -> Any:
            return cls.ATTR

    assert not VictimEq._eq_cls__check(Victim0)
    assert not VictimEq._eq_cls__check(Victim0())
    assert not VictimEq()._eq_cls__check(Victim0())
    assert not VictimEq()._eq_cls__check(Victim0)


def test__correct_usage():
    class VictimEqBase(Nest_EqCls):
        ATTR1: Any
        ATTR2: Any
        @classmethod
        @property
        def _EQ_CLS__VALUE(cls) -> Any:
            return cls.ATTR1 + cls.ATTR2

    class VictimEq1(VictimEqBase):
        ATTR1 = 1
        ATTR2 = 2

    class VictimEq2(VictimEqBase):
        ATTR1 = 2
        ATTR2 = 1

    assert VictimEq1 != VictimEq2
    assert VictimEq1 != VictimEq2()
    assert VictimEq1() != VictimEq2()
    assert VictimEq1() != VictimEq2

    assert VictimEq1._eq_cls__check(VictimEq2)
    assert VictimEq1._eq_cls__check(VictimEq2())
    assert VictimEq1()._eq_cls__check(VictimEq2())
    assert VictimEq1()._eq_cls__check(VictimEq2)

    VictimEq1().ATTR1 = 100
    assert VictimEq1 != VictimEq2
    assert VictimEq1 != VictimEq2()
    assert VictimEq1() != VictimEq2()
    assert VictimEq1() != VictimEq2

    assert VictimEq1._eq_cls__check(VictimEq2)
    assert VictimEq1._eq_cls__check(VictimEq2())
    assert VictimEq1()._eq_cls__check(VictimEq2())
    assert VictimEq1()._eq_cls__check(VictimEq2)

    VictimEq1.ATTR1 = 100
    assert VictimEq1 != VictimEq2
    assert VictimEq1 != VictimEq2()
    assert VictimEq1() != VictimEq2()
    assert VictimEq1() != VictimEq2

    assert not VictimEq1._eq_cls__check(VictimEq2)
    assert not VictimEq1._eq_cls__check(VictimEq2())
    assert not VictimEq1()._eq_cls__check(VictimEq2())
    assert not VictimEq1()._eq_cls__check(VictimEq2)


def test__staticmethod():
    class VictimEqBase(Nest_EqCls):
        ATTR1: Any
        ATTR2: Any
        @classmethod
        @property
        def _EQ_CLS__VALUE(cls) -> Any:
            return cls.ATTR1 + cls.ATTR2

    class VictimEq1(VictimEqBase):
        ATTR1 = 1
        ATTR2 = 2

    class VictimEq2(VictimEqBase):
        ATTR1 = 2
        ATTR2 = 1

    assert Nest_EqCls._eq_classes__check(VictimEq1, VictimEq2)
    assert Nest_EqCls._eq_classes__check(VictimEq1, VictimEq2())
    assert Nest_EqCls._eq_classes__check(VictimEq1(), VictimEq2())
    assert Nest_EqCls._eq_classes__check(VictimEq1(), VictimEq2)


# =====================================================================================================================



from base_aux.versions.m2_version import *
from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m3_exceptions import *

from base_aux.tables.m1_table_obj import TableObj


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="schema, _EXPECTED",
    argvalues=[
        (dict(ATC=1), False),
        (dict(ATC=[1,]), True),
        (dict(ATC=[1,], PTB=[1,]), True),
        (dict(ATC=[1,], PTB=[1,]), True),
        (dict(ATC=[1,], PTB=[]), False),
    ]
)
def test__LAMBDA_TRICK(schema, _EXPECTED):
    # INCORRECT! - too sophisticated!
    Lambda(lambda **kwargs: TableObj(**kwargs), **schema).check_raised__assert(not _EXPECTED)

    # CORRECT! - dont use any lambda params! do direct application
    Lambda(lambda: TableObj(**schema)).check_raised__assert(not _EXPECTED)


# =====================================================================================================================



from typing import *

from base_aux.aux_attr.m1_annot_attr1_aux import *
from base_aux.aux_iter.m1_iter_aux import *


# =====================================================================================================================
class NestContains_AttrIcNoPrivate:
    """
    GOAL
    ----
    apply str/repr for show attrs names+values

    CAREFUL
    -------
    dont use in Nest* classes - it can used only in FINALs!!! cause it can have same or meaning is not appropriate!
    """
    def __contains__(self, item: Any) -> bool:
        return IterAux([*AttrAux_Existed(self).iter__names_filter__not_private()]).item__get_original(item) not in [NoValue, None]
        # return AttrAux_Existed(self).name_ic__check_exists(item)


# =====================================================================================================================
class NestContains_AttrIcNotHidden:
    """
    GOAL
    ----
    apply str/repr for show attrs names+values

    CAREFUL
    -------
    dont use in Nest* classes - it can used only in FINALs!!! cause it can have same or meaning is not appropriate!
    """
    def __contains__(self, item: Any):
        return IterAux([*AttrAux_Existed(self).iter__names_filter__not_hidden()]).item__get_original(item) not in [NoValue, None]


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *

from base_aux.base_values.m5_value_valid2_variants import *
from base_aux.base_values.m5_value_valid3_unit import *


# =====================================================================================================================
class Test__ValueVariants:
    @classmethod
    def setup_class(cls):
        pass
        cls.Victim = type("Victim", (ValueVariants,), {})

    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__val_def__reset(self):
        victim = self.Victim(value="var1", variants=["VAR1", "VAR2"])
        assert victim.VALUE == "VAR1"
        assert victim.VALUE != "VAR2"
        assert victim.VALUE_DEFAULT == "VAR1"

        victim.VALUE = "var2"
        assert victim.VALUE != "VAR1"
        assert victim.VALUE == "VAR2"
        assert victim.VALUE_DEFAULT == "VAR1"

        victim.value_update()
        assert victim.VALUE == "VAR1"
        assert victim.VALUE != "VAR2"
        assert victim.VALUE_DEFAULT == "VAR1"

    def test__double_objects(self):
        victim1 = self.Victim(value="var1", variants=["VAR1", "VAR11"])
        victim2 = self.Victim(value="var2", variants=["VAR2", "VAR22"])
        assert victim1.VALUE == "VAR1"
        assert victim2.VALUE == "VAR2"

        assert victim1.VALUE != "VAR11"
        assert victim2.VALUE != "VAR22"

        victim1.VALUE = "VAR11"
        victim2.VALUE = "VAR22"

        assert victim1.VALUE != "VAR1"
        assert victim2.VALUE != "VAR2"

        assert victim1.VALUE == "VAR11"
        assert victim2.VALUE == "VAR22"

        try:
            victim1.VALUE = "VAR2"
        except:
            pass
        else:
            assert False

    def test__case(self):
        victim = self.Victim(value="var1", variants=["VAR1", "VAR2"])
        assert victim.VALUE == "VAR1"
        assert str(victim) == "VAR1"

        try:
            victim = self.Victim(value="var1", variants=["VAR1", "VAR2"], ignorecase=False)
        except:
            pass
        else:
            assert False

    def test__types__None(self):
        victim = self.Victim(variants=["NONE", ])
        assert victim.VALUE == NoValue
        # assert str(victim) == "NONE"

        victim = self.Victim(value=None, variants=["NONE", ])
        assert victim.VALUE == "NONE"
        assert str(victim) == "NONE"

        victim = self.Victim(value="None", variants=[None, ])
        assert victim.VALUE is None
        assert str(victim) == "None"

    def test__types__int(self):
        victim = self.Victim(value=1, variants=["1", ])
        assert victim.VALUE == "1"
        assert str(victim) == "1"

        victim = self.Victim(value="1", variants=[1, ])
        assert victim.VALUE == 1
        assert str(victim) == "1"

    def test__cmp__same_obj(self):
        assert self.Victim(value=None, variants=["NONE", ]) == self.Victim(value="None", variants=["NONE", ])
        assert self.Victim(value="NONE", variants=["NONE", ]) == self.Victim(value="None", variants=["NONE", ])
        assert self.Victim(value=None, variants=["None", ]) == self.Victim(value=None, variants=["NONE", ])

        assert self.Victim(value=None, variants=["None", ], ignorecase=False) != self.Victim(value=None, variants=["NONE", ])

    def test__cmp__simple_value(self):
        assert self.Victim(value=None, variants=["NONE", ]) == "NONE"
        assert self.Victim(value="NONE", variants=["NONE", ]) == "NONE"
        assert self.Victim(value=None, variants=["None", ]) == "NONE"
        assert self.Victim(value=None, variants=[None, ]) == "NONE"
        assert self.Victim(value=None, variants=[None, ], ignorecase=False) != "NONE"
        assert self.Victim(value=None, variants=[None, ], ignorecase=False) == "None"

    def test__contain(self):
        victim = self.Victim(variants=["NONE", ])
        assert None in victim
        assert "None" in victim
        assert "NONE" in victim

    def test__len(self):
        assert len(self.Victim(variants=[0, ])) == 1
        assert len(self.Victim(variants=[0, 1])) == 2

    def test__iter(self):
        assert list(self.Victim(variants=[0, ])) == [0, ]
        assert list(self.Victim(variants=[0, 1])) == [0, 1, ]

    def test__getitem(self):
        assert self.Victim(variants=[0, 1])[0] == 0
        assert self.Victim(variants=[0, 1])[1] == 1
        assert self.Victim(variants=[0, 1])[-1] == 1

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source1, obj2, _EXPECTED",
        argvalues=[
            (0, ValueVariants(0, variants=[0, 1]), True),
            (1, ValueVariants(0, variants=[0, 1]), False),
            (2, ValueVariants(0, variants=[0, 1]), False),

            ("0", ValueVariants(0, variants=[0, 1]), True),
            ("00", ValueVariants(0, variants=[0, 1]), False),

            (ValueUnit(0), ValueVariants(0, variants=[0, 1]), True),
        ]
    )
    def test__cmp_objs__value(self, source1, obj2, _EXPECTED):
        func_link = lambda: source1 == obj2
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source1, obj2, _EXPECTED",
        argvalues=[
            (0, ValueVariants(variants=[0, 1]), True),
            (1, ValueVariants(variants=[0, 1]), True),
            (2, ValueVariants(variants=[0, 1]), False),

            ("0", ValueVariants(variants=[0, 1]), True),
            ("00", ValueVariants(variants=[0, 1]), False),

            (ValueVariants(variants=[0, 1]), ValueUnit(0), True),
            (ValueUnit(0), ValueVariants(variants=[0, 1]), True),
        ]
    )
    def test__cmp_objs__no_value(self, source1, obj2, _EXPECTED):
        func_link = lambda: source1 == obj2
        Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================
def test__0():
    vu = ValueUnit(0)
    vv = ValueVariants(variants=[0, 1])

    assert vv == vu
    assert vu == vv


# =====================================================================================================================



from base_aux.base_nest_dunders.m6_eq1_attrs import *
from base_aux.aux_attr.m4_kits import *


# =====================================================================================================================
class VictimBase:
    o = 1
    _h = 1
    __p = 1


class VictimEqNotPivate(VictimBase, NestEq_AttrsNotPrivate):
    pass


class VictimEqNotHidden(VictimBase, NestEq_AttrsNotHidden):
    pass


# =====================================================================================================================
def test__base():
    victim = AttrKit_Blank()
    assert victim == VictimBase
    assert victim == VictimBase()
    assert VictimBase == victim
    assert VictimBase() == victim

    victim = AttrKit_Blank(o=1)
    assert victim == VictimBase
    assert victim == VictimBase()
    assert VictimBase == victim
    assert VictimBase() == victim

    victim = AttrKit_Blank(o=111)
    assert victim != VictimBase
    assert victim != VictimBase()
    assert VictimBase != victim
    assert VictimBase() != victim

    victim = AttrKit_Blank(o2=111)
    assert victim != VictimBase
    assert victim != VictimBase()
    assert VictimBase != victim
    assert VictimBase() != victim


def test__nested_p():
    victim = AttrKit_Blank(o=1, _h=1, __p=1)
    assert victim == VictimEqNotPivate
    assert victim == VictimEqNotPivate()
    assert VictimEqNotPivate == victim
    assert VictimEqNotPivate() == victim

    assert victim == VictimEqNotHidden
    assert victim == VictimEqNotHidden()
    assert VictimEqNotHidden == victim
    assert VictimEqNotHidden() == victim


def test__nested_h():
    victim = AttrKit_Blank(o=1, _h=1, __p=111)
    assert victim == VictimEqNotPivate
    assert victim == VictimEqNotPivate()
    assert VictimEqNotPivate == victim
    assert VictimEqNotPivate() == victim

    assert victim == VictimEqNotHidden
    assert victim == VictimEqNotHidden()
    assert VictimEqNotHidden == victim
    assert VictimEqNotHidden() == victim


def test__nested_o():
    victim = AttrKit_Blank(o=1, _h=111, __p=111)
    assert victim == VictimEqNotPivate
    assert victim == VictimEqNotPivate()
    assert VictimEqNotPivate == victim
    assert VictimEqNotPivate() != victim    # NOTE!

    assert victim == VictimEqNotHidden
    assert victim == VictimEqNotHidden()
    assert VictimEqNotHidden == victim
    assert VictimEqNotHidden() == victim


def test__nested_():
    victim = AttrKit_Blank(o=111, _h=111, __p=111)
    assert victim != VictimEqNotPivate
    assert victim != VictimEqNotPivate()
    assert VictimEqNotPivate != victim
    assert VictimEqNotPivate() != victim

    assert victim != VictimEqNotHidden
    assert victim != VictimEqNotHidden()
    assert VictimEqNotHidden != victim
    assert VictimEqNotHidden() != victim


# =====================================================================================================================



from base_aux.base_nest_dunders.m1_init1_args_kwargs import *
from base_aux.base_nest_dunders.m3_bool import *
from base_aux.base_nest_dunders.m3_calls import *
from base_aux.base_values.m3_exceptions import *
from base_aux.base_values.m4_primitives import *


# =====================================================================================================================
@final
class EqArgs(NestInit_Args_Implicit, NestCall_Resolve, NestBool_Resolve):
    """
    GOAL
    ----
    return True if all Args equal (with first arg)

    SPECIALLY CREATED FOR
    ---------------------
    DictDiff to cmp all elements in list with each other
    """
    def resolve(self) -> bool | NoReturn:
        # only one chans for Raise - not enough count --------------
        if len(self.ARGS) < 2:
            msg = f"need at least 2 args {self.ARGS=}"
            raise Exc__WrongUsage(msg)

        arg_0 = self.ARGS[0]

        # any raise on comparisons - return False! --------------
        for arg_next in self.ARGS[1:]:
            try:
                if arg_0 != arg_next:
                    return False
            except:
                return False

        return True

    def __eq__(self, other: Any | bool) -> bool | NoReturn:
        if other:   # True
            result = bool(self.resolve())
        else:      # False
            result = bool(self.resolve())

        return result


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *


# =====================================================================================================================
class Base_Signals(QObject):
    """
    see tests for usage examples

    RULES_LINE
    -----
    1. try connect and emit it in other classes!
    2. PARAMS
        - use always exact params COUNT as DEFINED in signal
        - use always defined TYPE!
        in this way it will be clearly/direct/expected behaviour and this is only best way!!!
        other variants - could work unobvious and very unexpected (see tests as examples)!!!

    EXCEPTIONS
    ----------
    in PyQt you will not get an Exc! just stop the program!
    1. if your code stops on EMIT() - see the exact target! don't forget use mandatory params!
    """
    pass
    # TYPICAL SIGNALS
    # signal__finished = pyqtSignal()
    # signal__stop = pyqtSignal()
    # signal__send_value = pyqtSignal(int)
    # signal__send_values = pyqtSignal(int, int)


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_text.m5_re2_attemps import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, pat, _EXPECTED",
    argvalues=[
        ("abc123", r"hello", Exception),
        ("abc123", r"\d+", "123"),
        ("abc123", r"\d(\d)\d", "2"),
        ("abc123", r"\d(\d)(\d)", ("2", "3")),
    ]
)
def test__result_get_from_match(source, pat, _EXPECTED):
    match = re.search(pat, source)
    func_link = Base_ReAttempts._result__get_from_match
    Lambda(func_link, match).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Test__re:
    @pytest.mark.parametrize(
        argnames="source, attempts, _EXPECTED",
        argvalues=[
            # GROUPS-NO --------
            ("aaa111bbb222", [r"\d+", ], [
                None, [],
                None, [],
                "111", ["111", ],
                ["111", "222"], ["111", "222", ],
                "aaabbb", "aaabbb",
            ]),
            ("aaa111bbb222", [r"\d+", r"\d+"], [        # same rexp
                None, [],
                None, [],
                "111", ["111", "111"],
                ["111", "222"], ["111", "222", "111", "222"],
                "aaabbb", "aaabbb",
            ]),
            ("aaa111bbb222", [r"\d+", r"\D+"], [
                "aaa", ["aaa", ],
                None, [],
                "111", ["111", "aaa"],
                ["111", "222"], ["111", "222", "aaa", "bbb"],
                "aaabbb", "",
            ]),

            # groups-ONE --------
            ("aaa111bbb222", [r"b(\d+)", ], [
                None, [],
                None, [],
                "222", ["222", ],
                ["222"], ["222", ],
                "aaa111bb", "aaa111bb",
            ]),
            # groups-several --------
            ("aaa111bbb222", [r"(\d+)\D+(\d+)", ], [
                None, [],
                None, [],
                ("111", "222"), [("111", "222"), ],
                [("111", "222")], [("111", "222"), ],
                "aaa", "aaa",
            ]),

            # groups-several --------
            ("aaa111bbb222", [RExp(r"(\D+)(\d+)", sub=r"\2\1"), ], [
                ("aaa", "111"), [("aaa", "111"), ],
                None, [],
                ("aaa", "111"), [("aaa", "111"), ],
                [("aaa", "111"), ("bbb", "222")], [("aaa", "111"), ("bbb", "222"), ],
                "111aaa222bbb", "111aaa222bbb",
            ]),
        ]
    )
    def test__match(self, source, attempts, _EXPECTED):
        Lambda(ReAttemptsFirst(*attempts).match, source).check_expected__assert(_EXPECTED[0])
        Lambda(ReAttemptsAll(*attempts).match, source).check_expected__assert(_EXPECTED[1])

        Lambda(ReAttemptsFirst(*attempts).fullmatch, source).check_expected__assert(_EXPECTED[2])
        Lambda(ReAttemptsAll(*attempts).fullmatch, source).check_expected__assert(_EXPECTED[3])

        Lambda(ReAttemptsFirst(*attempts).search, source).check_expected__assert(_EXPECTED[4])
        Lambda(ReAttemptsAll(*attempts).search, source).check_expected__assert(_EXPECTED[5])

        Lambda(ReAttemptsFirst(*attempts).findall, source).check_expected__assert(_EXPECTED[6])
        Lambda(ReAttemptsAll(*attempts).findall, source).check_expected__assert(_EXPECTED[7])

        Lambda(ReAttemptsFirst(*attempts).sub, source).check_expected__assert(_EXPECTED[8])
        Lambda(ReAttemptsAll(*attempts).sub, source).check_expected__assert(_EXPECTED[9])
        Lambda(ReAttemptsFirst(*attempts).delete, source).check_expected__assert(_EXPECTED[8])
        Lambda(ReAttemptsAll(*attempts).delete, source).check_expected__assert(_EXPECTED[9])

    @pytest.mark.parametrize(
        argnames="source, attempts, _EXPECTED",
        argvalues=[
            ("aaa111bbb222", [RExp(r"(\D+)(\d+)", sub=r"\2\1"), ], [
                "111aaa222bbb", "111aaa222bbb",
            ]),
            ("aaa111bbb222", [RExp(r"(\D+)(\d+)", sub=r"\2\1", scount=1), ], [
                "111aaabbb222", "111aaabbb222",
            ]),

            ("aaa111bbb222", [RExp(r"(\D+)(\d+)", sub=r"\2\1"), RExp(r"(\D+)(\d+)", sub=r"\2\1")], [
                "111aaa222bbb", "111222aaabbb",
            ]),
        ]
    )
    def test__sub(self, source, attempts, _EXPECTED):
        Lambda(ReAttemptsFirst(*attempts).sub, source).check_expected__assert(_EXPECTED[0])
        Lambda(ReAttemptsAll(*attempts).sub, source).check_expected__assert(_EXPECTED[1])


# =====================================================================================================================



import pytest

from base_aux.buses.m1_serial1_client import *
from base_aux.buses.m1_serial2_client_derivatives import *


# =====================================================================================================================
@pytest.mark.skipif(condition=not SerialClient.addresses_paired__detect(), reason="NO addresses_paired__detect====FIXME!!!")
class Test__Paired:
    Victim: type[SerialClient_FirstFree_Paired]
    victim: SerialClient_FirstFree_Paired

    SETUP_CLS__CONNECT: bool = False
    TEARDOWN__DISCONNECT: bool = True

    @classmethod
    def setup_class(cls):
        print(SerialClient_FirstFree_Paired.addresses_paired__detect())

        class Victim(SerialClient_FirstFree_Paired):
            LOG_ENABLE = True

            # def address__validate(self) -> Union[bool, NoReturn]:
            #     return self.write_read_line_last("echo") == "echo"

        cls.Victim = Victim
        cls.victim = cls.Victim()

        class Victim(SerialClient_FirstFree_Shorted):
            RAISE_CONNECT = False

        cls.Victim = Victim
        cls.victim = cls.Victim()
        if cls.SETUP_CLS__CONNECT and not cls.victim.connect():
            msg = f"[ERROR] not found PORTs PAIRED"
            print(msg)
            raise Exception(msg)

    @classmethod
    def teardown_class(cls):
        pass
        if hasattr(cls, "victim") and cls.victim:
            cls.victim._addresses__release()
            cls.victim.disconnect()

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass
        if self.TEARDOWN__DISCONNECT and hasattr(self, "victim") and self.victim:
            self.victim._addresses__release()
            self.victim.disconnect()

    # -----------------------------------------------------------------------------------------------------------------
    def test__ADDRESS__PAIRED(self):
        # self.victim.disconnect()

        assert self.victim.addresses_paired__count() > 0    # HERE YOU NEED CONNECT/CREATE/COMMUTATE ONE PAIR!
        print(f"{self.victim.ADDRESSES__PAIRED=}")

        assert self.victim.connect() is True

        addr_paired = self.victim.address_paired__get()
        assert addr_paired is not None

        victim_pair = SerialClient()
        assert victim_pair.connect(address=addr_paired)

        assert self.victim._write("LOAD1")
        assert victim_pair.read_line() == "LOAD1"

        assert victim_pair._write("LOAD2")
        assert self.victim.read_line() == "LOAD2"


# =====================================================================================================================



from typing import *

from base_aux.base_values.m3_exceptions import Exc__Incompatible


# =====================================================================================================================
class TableObj:
    """
    GOAL
    ----
    create table of objects (lines+columns)
    all lines are same len
    """
    schema: dict[str, list[Any] | Collection]   # Collection if for universal

    def __init__(
            self,
            **schema: list[Any] | Collection,
    ) -> None | NoReturn:
        if not self._validate_schema(schema):
            msg = f"[{self.__class__.__name__}]countColumns/{schema=}"
            raise Exc__Incompatible(msg)

        self.schema = schema
        try:
            self.__count_columns: int = len(list(self.schema.values())[0])
        except:
            self.__count_columns: int = 0

    @classmethod
    def _validate_schema(cls, schema: dict[str, list[Any]]) -> bool:
        """
        GOAL
        ----
        just check the schema is correct!
        mainly by counts!
        """
        len_expect = None
        if not isinstance(schema, dict):
            # just in case!
            msg = f"[{cls.__name__}]/{schema=}"
            print(msg)
            return False

        for name, line in schema.items():
            if not isinstance(name, str):
                # just in case!
                msg = f"[{cls.__name__}]/{name=}"
                print(msg)
                return False

            try:
                len_cur = len(line)
            except Exception as exc:
                print(f"{exc!r}")
                return False

            # init expected
            if len_expect is None:
                len_expect = len_cur

            if len_expect != len_cur:
                msg = f"[{cls.__name__}]/{name=}/{len_expect=}/{len_cur=}"
                print(msg)
                return False

        return True

    def count_lines(self) -> int:
        return len(self.schema)

    def count_columns(self) -> int:
        return self.__count_columns

    def __getitem__(self, item: str) -> list[Any] | NoReturn:
        return self.schema[item]

    def __getattr__(self, item: str) -> list[Any] | NoReturn:   # FIXME: deprecate???
        # if item.startswith("__") and item.endswith("__"):
        #     raise KeyError()
        return self.schema[item]

    @property
    def __name__(self) -> str:
        """
        GOAL
        ----
        just for suppression internal unexpected error when using tests for TableColumn!
        it appears only in there!
        """
        return self.__class__.__name__


# =====================================================================================================================



import pathlib

from base_aux.aux_text.m1_text_aux import TextAux
from base_aux.base_types.m0_static_typing import TYPING
from base_aux.path2_file.m2_file import FileAux

from base_aux.aux_text.m0_patterns import *


# =====================================================================================================================
class TextFile(FileAux, TextAux):
    """
    GOAL
    ----
    same as FileAux but with TextAux methods applied inplace!
    """
    def __init__(
            self,
            filepath: TYPING.PATH_DRAFT = None,
            text: TYPING.STR_DRAFT = None,
            # *args, **kwargs
    ) -> None | NoReturn:
        # super().__init__(*args, **kwargs)     # NOTE: dont use here!??? because init is overWriten here!

        if filepath is not None:
            self.FILEPATH = pathlib.Path(filepath)
        if self.check_exists() and not self.FILEPATH.is_file():
            raise Exc__Incompatible_Data(f"{self.FILEPATH=}")

        if text is not None:
            self.TEXT = str(text)
        else:
            if self.check_exists():
                self.read__text()


# =====================================================================================================================



import time
from typing import *
import asyncio


# =====================================================================================================================
class Sleep:
    """
    GOAL
    ----
    just a primitive func for tests or other purpose!
    """
    DEF_SEC: float = 1
    sec: float

    def __init__(self, sec: float = None):
        if sec is not None:
            self.sec = sec
        else:
            self.sec = self.DEF_SEC

    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.sec})"

    def __repr__(self) -> str:
        return str(self)

    # -----------------------------------------------------------------------------------------------------------------
    def echo(self, echo: Any = None, *args, **kwargs) -> Any:
        time.sleep(self.sec)
        return echo

    def NONE(self, *args, **kwargs) -> None:
        # NOTE: why used uppercase!? - cause "raise" name here will not be appropriate!
        return self.echo(echo=None, *args, **kwargs)

    def TRUE(self, *args, **kwargs) -> bool:
        return self.echo(echo=True, *args, **kwargs)

    def FALSE(self, *args, **kwargs) -> bool:
        return self.echo(echo=False, *args, **kwargs)

    def EXC(self, *args, **kwargs) -> Exception:
        return self.echo(echo=Exception("Sleep.EXC"), *args, **kwargs)

    def RAISE(self, *args, **kwargs) -> NoReturn:
        self.echo()
        raise Exception("Sleep.RAISE")

    # -----------------------------------------------------------------------------------------------------------------
    async def aio_echo(self, echo: Any = None, *args, **kwargs) -> Any:
        await asyncio.sleep(self.sec)
        return echo

    async def aio_NONE(self, *args, **kwargs) -> None:
        return await self.aio_echo(echo=None, *args, **kwargs)

    async def aio_TRUE(self, *args, **kwargs) -> bool:
        return await self.aio_echo(echo=True, *args, **kwargs)

    async def aio_FALSE(self, *args, **kwargs) -> bool:
        return await self.aio_echo(echo=False, *args, **kwargs)

    async def aio_EXC(self, *args, **kwargs) -> Exception:
        return await self.aio_echo(echo=Exception("Sleep.EXC"), *args, **kwargs)

    async def aio_RAISE(self, *args, **kwargs) -> NoReturn:
        await self.aio_echo()
        raise Exception("Sleep.RAISE")


# =====================================================================================================================
class Base_SleepAw(Sleep):
    async def start(self) -> Any | NoReturn:
        raise NotImplementedError()

    def __await__(self) -> None | NoReturn:
        result = yield from self.start().__await__()
        return result


# ---------------------------------------------------------------------------------------------------------------------
class SleepAwNone(Base_SleepAw):
    async def start(self) -> None:
        return await self.aio_NONE()


class SleepAwTrue(Base_SleepAw):
    async def start(self) -> bool:
        return await self.aio_TRUE()


class SleepAwFalse(Base_SleepAw):
    async def start(self) -> bool:
        return await self.aio_FALSE()


class SleepAwExc(Base_SleepAw):
    async def start(self) -> Exception:
        return await self.aio_EXC()


class SleepAwRaise(Base_SleepAw):
    async def start(self) -> NoReturn:
        return await self.aio_RAISE()


# =====================================================================================================================



from base_aux.aux_eq.m2_eq_aux import *


# =====================================================================================================================
class NestInit_AnnotsAttr_ByArgs:
    def __init__(self, *args: Any, **kwargs: TYPING.KWARGS_FINAL) -> None | NoReturn:
        AttrAux_AnnotsAll(self).sai__by_args(*args)
        super().__init__(**kwargs)


# =====================================================================================================================
class NestInit_AnnotsAttr_ByKwargs:
    """
    SPECIALLY CREATED FOR
    ---------------------
    HtmlTagParser to init attrs only by kwargs
    """
    def __init__(self, *args: Any, **kwargs: TYPING.KWARGS_FINAL) -> None | NoReturn:
        AttrAux_AnnotsAll(self).sai__by_kwargs(**kwargs)
        super().__init__(*args)


# =====================================================================================================================
class NestInit_AnnotsAttr_ByArgsKwargs(NestInit_AnnotsAttr_ByArgs, NestInit_AnnotsAttr_ByKwargs):     # NOTE: dont create AnnotsOnly/AttrsOnly! always use this class!
    """
    NOTE
    ----
    1. for more understanding application/logic use annots at first place! and dont mess them. keep your code clear!
        class Cls(NestInit_AnnotsAttr_ByArgsKwargs):
            A1: Any
            A2: Any
            A3: Any = 1
            A4: Any = 1

    2. mutable values are acceptable!!!

    GOAL
    ----
    init annots/attrs by params in __init__

    LOGIC
    -----
    ARGS
        - used for ANNOTS ONLY - used as values! not names!
        - inited first without Kwargs sense
        - if args less then annots - no matter
        - if args more then annots - no matter+no exc
        - if kwargs use same keys - it will overwrite by kwargs (args set first)
    KWARGS
        - used for both annots/attrs (annots see first)
        - if not existed in Annots and Attrs - create new!
    """
    pass


# ---------------------------------------------------------------------------------------------------------------------
# class NestInit_AnnotsAttrByKwArgsIc(NestInit_AnnotsAttr_ByArgsKwargs, NestGSAI_AttrAnycase):   # IC - IS NOT WORKING!!!
#     """
#     SAME AS - 1=parent
#     -------
#     but attrs access will be IgnoreCased
#     """
#     pass


# =====================================================================================================================
if __name__ == '__main__':
    pass


# =====================================================================================================================



from base_aux.valid.m1_valid_base import *
from base_aux.translator.m1_translator import Translator
from base_aux.pyqt.m3_tm import Base_TableModel
from base_aux.breeders.m1_breeder_str2_stack import *
from base_aux.breeders.m1_breeder_str1_series import *
from base_aux.pyqt.m0_static import *


# =====================================================================================================================
class TableModel_Devs(Base_TableModel):
    DATA: "TpManager"
    HEADERS: "Headers"
    HTRANSLATOR: Translator

    # AUX -------------------------------------------
    open__settings: Optional[bool] = None

    def __init__(self, data: "TpManager" = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reinit(data)

    def reinit(self, data: "TpManager" = None) -> None:
        """
        GOAL
        ----
        separate inition!
        """
        if data is not None:
            self.DATA = data

        class Headers(BreederStrStack):
            NAME: int = 0
            DEVICE: BreederStrSeries = BreederStrSeries(None, self.DATA.STAND.DEV_LINES.COUNT_COLUMNS)

        class HTRus:
            NAME: str = "Имя"

        self.HEADERS = Headers()
        self.HTRANSLATOR = Translator(HTRus)

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return len(self.DATA.STAND.DEV_LINES)

    def columnCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return self.HEADERS.count()

    # def headerData(self, section: Any, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> str:
    #     if role == Qt.DisplayRole:
    #         # ------------------------------
    #         if orientation == Qt.Horizontal:
    #             return self.HEADERS[section]
    #
    #         # ------------------------------
    #         if orientation == Qt.Vertical:
    #             return str(section + 1)

    # def flags(self, index: QModelIndex) -> int:
    #     # PREPARE -----------------------------------------------------------------------------------------------------
    #     col = index.column()
    #     row = index.row()
    #
    #     dev_group_name: str | None = None
    #     try:
    #         dev_group_name = self.DATA.STAND.DEV_LINESS.names()[row]
    #     except:
    #         pass
    #
    #     row_is_summary: bool = dev_group_name is None
    #
    #     # -------------------------------------------------------------------------------------------------------------
    #     flags = super().flags(index)
    #
    #     if row_is_summary:
    #         pass
    #
    #     elif col in [self.HEADERS.DEVICE, self.HEADERS.ADDRESS] or col in self.HEADERS.CONNECTED:
    #         flags |= Qt.ItemIsUserCheckable
    #         # flags |= Qt.ItemIsSelectable
    #     else:
    #         # flags -= Qt.ItemIsSelectable
    #         pass
    #
    #     # clear SELECTABLE ---------
    #     return flags

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()

        dev_group_name: str | None = None
        dev_inst: Any | None = None

        try:
            # print(f"{self.DATA.STAND.DEV_LINES.names()=}")
            dev_group_name = self.DATA.STAND.DEV_LINES.names()[row]
        except:
            pass

        if dev_group_name and col in self.HEADERS.DEVICE:
            index = col - self.HEADERS.DEVICE.START_OUTER
            try:
                dev_inst = self.DATA.STAND.DEV_LINES[dev_group_name][index]
            except:
                dev_inst = self.DATA.STAND.DEV_LINES[dev_group_name]

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.DisplayRole:
            if col == self.HEADERS.NAME:
                # return f"{tc_cls.NAME}\n{tc_cls.DESCRIPTION}"
                return f"{dev_group_name}"
            if col in self.HEADERS.DEVICE:
                if dev_inst and dev_inst.DEV_FOUND:
                    return dev_inst.ADDRESS
                else:
                    return "-"

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.TextAlignmentRole:
            """
            VARIANTS ALIGN
            --------------
            not exists NAME!!!} = 0         # (LEFT+TOP) [[[[[[[[DEFAULT IS [LEFT+TOP]]]]]]]]]
            
            AlignLeft=AlignLeading = 1      # LEFT(+TOP)
            AlignRight=AlignTrailing = 2    # RIGHT(+TOP)

            AlignTop = 32       # TOP(+LEFT)
            AlignBottom = 64    # BOT(+LEFT)

            AlignHCenter = 4    # HCENTER(+TOP)
            AlignVCenter = 128  # VCENTER(+LEFT)
            AlignCenter = 132   # VCENTER+HCENTER

            # =====MAYBE DID NOT FIGURED OUT!!!
            AlignAbsolute = 16      # (LEFT+TOP) == asDEFAULT
            AlignBaseline = 256     # (LEFT+TOP) == asDEFAULT

            AlignJustify = 8        # (LEFT+TOP) == asDEFAULT

            AlignHorizontal_Mask = 31   # TOP+RIGHT
            AlignVertical_Mask = 480    # LEFT+VCENTER
            """
            if col == self.HEADERS.NAME:
                return ALIGNMENT.CL
            else:
                return ALIGNMENT.C

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.BackgroundColorRole:
            if col in self.HEADERS.DEVICE:
                if dev_inst:
                    if dev_inst.DEV_FOUND:
                        return QColor('#50FF50')
                    else:
                        return QColor('#FF5050')


# =====================================================================================================================



import pytest

from base_aux.buses.m1_serial1_client import *
from base_aux.buses.m1_serial2_client_derivatives import *


# =====================================================================================================================
@pytest.mark.skip
class Test__ATC:
    victim: SerialClient

    @classmethod
    def setup_class(cls):
        pass

        class Atc_SerialClient(SerialClient_FirstFree_AnswerValid):
            LOG_ENABLE = True
            RAISE_CONNECT = False
            BAUDRATE = 115200

            # # ATC ----------------------------
            PREFIX = "ATC:03:"
            EOL__SEND = b"\r"

            def address__validate(self) -> bool:
                return self.write_read__last_validate("get name", "ATC 03")

        cls.victim = Atc_SerialClient()

    # @classmethod
    # def teardown_class(cls):
    #     if cls.victim:
    #         cls.victim.disconnect()
    #
    # def setup_method(self, method):
    #     pass

    def teardown_method(self, method):
        pass
        if self.victim:
            self.victim.disconnect()

    # -----------------------------------------------------------------------------------------------------------------
    def test__1(self):
        assert self.victim.connect()
        print(f"{self.victim.connect()=}")
        # print(f"{self.victim.addresses_system__detect()=}")
        print(f"{self.victim.ADDRESS=}")
        assert self.victim.address_check__resolved()


# =====================================================================================================================
@pytest.mark.skip
class Test__PTB(Test__ATC):
    @classmethod
    def setup_class(cls):
        pass

        class Ptb_SerialClient(SerialClient_FirstFree_AnswerValid):
            LOG_ENABLE = True
            RAISE_CONNECT = False
            BAUDRATE = 115200

            # PTB ----------------------------
            PREFIX = "PTB:01:"
            EOL__SEND = b"\n"

            def address__validate(self) -> bool:
                return self.write_read__last_validate("get name", "PTB") and self.write_read__last_validate("get addr", "01")

        cls.victim = Ptb_SerialClient()


# =====================================================================================================================



from base_aux.base_types.m0_static_typing import TYPING
from base_aux.base_nest_dunders.m1_init1_source import *
from base_aux.base_types.m1_type_aux import TypeAux

from base_aux.aux_argskwargs.m1_argskwargs import ArgsKwargs
from base_aux.base_types.m0_static_types import *


# =====================================================================================================================
@final
class ArgsKwargsAux(NestInit_Source):
    SOURCE: TYPING.ARGS_DRAFT | TYPING.KWARGS_DRAFT

    def resolve_args(self) -> TYPING.ARGS_FINAL:     # REPLACING for args__ensure_tuple
        if isinstance(self.SOURCE, ArgsKwargs):
            return self.SOURCE.ARGS
        elif TypeAux(self.SOURCE).check__elementary_collection():
            return tuple(self.SOURCE)
        else:
            return (self.SOURCE,)

    def resolve_kwargs(self) -> TYPING.KWARGS_FINAL | NoReturn:
        if isinstance(self.SOURCE, ArgsKwargs):
            return self.SOURCE.KWARGS
        elif not self.SOURCE:
            return {}
        else:
            return dict(self.SOURCE)


# =====================================================================================================================



from base_aux.aux_attr.m1_annot_attr1_aux import *
from base_aux.base_enums.m2_enum1_adj import *


# =====================================================================================================================
class Base_AttrDictDumping(NestInit_Source, NestCall_Resolve):
    """
    GOAL
    ----
    separate object to make only one thing - dumping attrs final state
    for exploring objects

    SPECIALLY CREATED FOR
    ---------------------
    replace ObjectInfo State with attrsGroups to simple flat name-value
    for make simplified comparing several states
    """
    SOURCE: Any
    SKIP_NAMES: tuple[str | Base_EqValid, ...]

    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ATTRS_EXISTED
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED

    def __init__(self, source: Any = NoValue, *skip_names: str | Base_EqValid) -> None:
        super().__init__(source)
        self.SKIP_NAMES = skip_names

    def resolve(self) -> dict[str, Any | Exception] | NoReturn:
        # select -----
        result_obj = None

        if self._ATTRS_STYLE == EnumAdj_AttrAnnotsOrExisted.ATTRS_EXISTED:
            result_obj = AttrAux_Existed(self.SOURCE, *self.SKIP_NAMES)
        elif self._ATTRS_STYLE == EnumAdj_AttrAnnotsOrExisted.ANNOTS_ONLY:
            if self._ANNOTS_DEPTH == EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED:
                result_obj = AttrAux_AnnotsAll(self.SOURCE, *self.SKIP_NAMES)
            elif self._ANNOTS_DEPTH == EnumAdj_AnnotsDepthAllOrLast.LAST_CHILD:
                result_obj = AttrAux_AnnotsLast(self.SOURCE, *self.SKIP_NAMES)
        else:
            raise Exc__Incompatible_Data(f"{self._ATTRS_STYLE=}/{self._ANNOTS_DEPTH=}")

        # result -----
        return result_obj.dump_dict()


# ---------------------------------------------------------------------------------------------------------------------
@final
class AttrDictDumping_Existed(Base_AttrDictDumping):
    """
    NOTE
    ----
    main class! most used
    next derivatives is not useful i think!
    """
    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ATTRS_EXISTED
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED


@final
class AttrDictDumping_AnnotsAll(Base_AttrDictDumping):
    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ANNOTS_ONLY
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED


@final
class AttrDictDumping_AnnotsLast(Base_AttrDictDumping):
    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ANNOTS_ONLY
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.LAST_CHILD


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from base_aux.versions.m3_derivatives import Version_Python


# =====================================================================================================================
class Test__ReqCheckVersion_Python:
    # -----------------------------------------------------------------------------------------------------------------
    def test__cmp_str(self):
        assert Version_Python("1.2rc2.3").check_eq("1.02rc2.3") is True
        assert Version_Python("1.2rc2.3").check_eq("1.2rc2.33") is False

        assert Version_Python("1.2rc2.3").check_ne("1.02rc2.3") is False
        assert Version_Python("1.2rc2.3").check_ne("1.2rc2.33") is True

        assert Version_Python("1.2").check_le("1.2rc2.3") is True
        assert Version_Python("1.2").check_lt("1.2rc2.3") is True

        assert Version_Python("1.2").check_ge("1.2rc2.3") is False
        assert Version_Python("1.2").check_gt("1.2rc2.3") is False

    def test__py(self):
        assert Version_Python().check_eq("1.02rc2.3") is False
        assert Version_Python().check_gt("1.02rc2.3") is True

    def test__raise(self):
        # IF -----------------
        print("START"*100)
        try:
            Version_Python("1.2").raise_if__check_eq("1.02")
        except:
            pass
        else:
            assert False

        Version_Python("1.2").raise_if__check_eq("1.22")

        # IF NOT -----------------
        Version_Python("1.2").raise_if_not__check_eq("1.02")
        try:
            Version_Python("1.2").raise_if_not__check_eq("1.22")
        except:
            pass
        else:
            assert False

    # @pytest.mark.parametrize(
    #     argnames="args, _EXPECTED",
    #     argvalues=[
    #         (("1.2rc2.3", "1.2rc2.3"), True),
    #     ]
    # )
    # def test__inst__cmp__eq(self, args, _EXPECTED):
    #     func_link = lambda source1, source2: Version_Python().check_eq("1.02rc2.3")
    #     Lambda(func_link, args).assert_check(_EXPECTED)


# =====================================================================================================================



import math

from base_aux.base_types.m0_static_types import *
from base_aux.base_types.m2_info import *

from base_aux.base_nest_dunders.m7_cmp import *


# =====================================================================================================================
class ArithmApplyToAttr(NestCmp_GLET_Any):
    """
    GOAL
    ----
    translate all arithmetic operations to exact attribute (expected as number)!

    NOTE
    ----
    1. BE CAREFUL to compare FLOATS!!! it used Precision settings!

    2. MAINLY fom the beginning:
    always return Self for any operation (except direct int/float transform methods).
    So if you need exact NUMBER - just apply float() (int is incorrect!).

    MAYBE IN FUTURE:
    IF FUNC - return number, if operation - return Self.
    I'm not really know what is the best decision but in future it may be changed for real implementation

    CREATED SPECIALLY FOR
    ---------------------
    funcs.ValueUnit and further ising in Uart Responses

    BEST USAGE
    ----------
    see tests!

    REFERENCE
    ---------
    https://docs.python.org/3/library/operator.html
    """

    # SETTINGS --------------------------------------------------------------------------------------------------------
    NUMBER_ARITHM__ATTR: str = None     # DEFINE!!! name for ORIGINALVALUE
    NUMBER_ARITHM__PRECISION: int | None = 6

    # AUX -------------------------------------------------------------------------------------------------------------
    @property
    def NUMBER_ARITHM(self) -> Union[TYPES.NUMBER, NoReturn]:
        if not self.NUMBER_ARITHM__ATTR:
            raise Exc__NotExistsNotFoundNotCreated()

        result = getattr(self, self.NUMBER_ARITHM__ATTR)
        if TypeAux(result).check__callable_func_meth_inst():
            result = result()
        return result

    @NUMBER_ARITHM.setter
    def NUMBER_ARITHM(self, other) -> None | NoReturn:
        if not self.NUMBER_ARITHM__ATTR:
            raise Exc__NotExistsNotFoundNotCreated()

        if self.NUMBER_ARITHM__PRECISION is not None:
            other = round(other, self.NUMBER_ARITHM__PRECISION)

        setattr(self, self.NUMBER_ARITHM__ATTR, other)

    # precision --------------------------------------
    @classmethod
    def number__fix_precision(cls, source: Any, round_n: int | None = None) -> TYPES.NUMBER:
        source = float(source)
        if round_n is None:
            round_n = cls.NUMBER_ARITHM__PRECISION
        if round_n is not None:
            source = round(source, round_n)

        # final -------
        source = cls.number__try_int_if_same(source)
        return source

    @classmethod
    def number__try_int_if_same(cls, source: Any) -> int | Any:
        """
        GOAL
        ----
        get int value if int()==float()

        USEFUL IDEAS
        ------------
        1. keep short value in str() when we have float value with insignificant zeros (like 1.0)
        2. correct comparing int with float
            assert 1 != 1.0
            assert int(1) == int(1.0)
        3. fix correct comparing strings
            assert int("1.0") == 1  # will get EXC!!!
            assert number__try_int_if_same("1.0") == 1  # will get True

        Created specially for
        ---------------------
        NumberArithm class
        """
        try:
            if int(float(source)) == float(source):
                source = int(float(source))
        except:
            pass
        return source

    @classmethod
    def number__get_string_no_zeros(cls, source: Any, round_n: int | None = None) -> str:
        source = cls.number__fix_precision(source, round_n)
        if isinstance(source, int):
            result = f"{source}"
        else:
            result= f"{source:f}".rstrip("0").rstrip(".")

        # final -------
        return result

    # CONVERT ---------------------------------------------------------------------------------------------------------
    # MAIN --------------------------------------
    def __int__(self) -> int | NoReturn:
        """
        CAUTION
        -------
        BE CAREFUL ABOUT ANY INT!!!
            int(1.999) == 1!!!!
        :return:
        """
        return int(self.NUMBER_ARITHM)

    def __float__(self) -> float:
        return float(self.NUMBER_ARITHM)

    def __hash__(self) -> int:
        return hash(self.NUMBER_ARITHM)

    # OTHER --------------------------------------
    def __bool__(self) -> bool:
        return bool(self.NUMBER_ARITHM)

    def __str__(self) -> str:
        return self.number__get_string_no_zeros(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.NUMBER_ARITHM})"

    # SIGN ------------------------------------------------------------------------------------------------------------
    def __pos__(self) -> Self:
        """
        для унарного плюса (+some_object)
        """
        return self

    def __neg__(self) -> Self:
        """
        для унарного минуса (-some_object)
        """
        self.NUMBER_ARITHM = - self.NUMBER_ARITHM
        return self

    def __abs__(self) -> Self:
        self.NUMBER_ARITHM = abs(self.NUMBER_ARITHM)
        return self

    # PARTS -----------------------------------------------------------------------------------------------------------
    def __round__(self, n=None) -> Self:
        self.NUMBER_ARITHM = round(self.NUMBER_ARITHM, n)
        return self

    def __floor__(self) -> Self:
        self.NUMBER_ARITHM = math.floor(self.NUMBER_ARITHM)
        return self

    def __ceil__(self) -> Self:
        self.NUMBER_ARITHM = math.ceil(self.NUMBER_ARITHM)
        return self

    def __trunc__(self) -> Self:
        self.NUMBER_ARITHM = math.trunc(self.NUMBER_ARITHM)
        return self

    # OTHER ===========================================================================================================
    def _other__get_float(self, other: Any) -> float | NoReturn:
        try:
            other = self.__class__(other)
        except:
            pass
        return float(other)

    # ARITHM ----------------------------------------------------------------------------------------------------------
    def __add__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM + self._other__get_float(other)
        return self

    def __sub__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM - self._other__get_float(other)
        return self

    def __mul__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM * self._other__get_float(other)
        return self

    # def __div__(self, other) -> Self:   # THERE IS NO SUCH MAGICMETH!!! use truediv!
    #     return self.NUMBER_ARITHM/other
    #     return self

    def __truediv__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM / self._other__get_float(other)
        return self

    def __floordiv__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM // self._other__get_float(other)
        return self

    def __mod__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM % self._other__get_float(other)
        return self

    def __divmod__(self, other) -> tuple[int, int | float]:
        return divmod(self.NUMBER_ARITHM, float(other))

    def __pow__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM ** self._other__get_float(other)
        return self

    # INLINE ----------------------------------------------------------------------------------------------------------
    def __iadd__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM + self._other__get_float(other)
        return self

    def __isub__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM - self._other__get_float(other)
        return self

    def __imul__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM * self._other__get_float(other)
        return self

    def __itruediv__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM / self._other__get_float(other)
        return self

    def __ifloordiv__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM // self._other__get_float(other)
        return self

    def __imod__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM % self._other__get_float(other)
        return self

    def __ipow__(self, other) -> Self:
        self.NUMBER_ARITHM = self.NUMBER_ARITHM ** self._other__get_float(other)
        return self

    # CMP -------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other) -> int | NoReturn:
        """
        do try to resolve Exceptions!!! sometimes it is ok to get it!!!

        RETURN
        ------
            1=self>other
            0=self==other
            -1=self<other
        """
        if float(self.NUMBER_ARITHM) == self._other__get_float(other) or other == float(self.NUMBER_ARITHM):
            return 0
        if float(self.NUMBER_ARITHM) > self._other__get_float(other) or other < float(self.NUMBER_ARITHM):
            return 1
        if float(self.NUMBER_ARITHM) < self._other__get_float(other) or other > float(self.NUMBER_ARITHM):
            return -1


# =====================================================================================================================



from base_aux.buses.m1_serial2_client_derivatives import *

from .models import *


# =====================================================================================================================
class Base_Device:
    NAME: str = None
    DESCRIPTION: str = None
    INDEX: int = None

    # PROPERTIES ------------------------------------------------------------------------------------------------------
    DEV_FOUND: bool | None = None

    SN: str = None
    FW: str = None
    MODEL: str = None

    # DUT --------------------
    SKIP: Optional[bool] = None

    DUT_SN: str = None
    DUT_FW: str = None
    DUT_MODEL: str = None

    # -----------------------------------------------------------------------------------------------------------------
    def SKIP_reverse(self) -> None:
        """
        this is only for testing purpose
        """
        self.SKIP = not bool(self.SKIP)

    # # CONNECT -------------------------------------------------------------------------------------------------------
    # def connect(self) -> bool:
    #     return True
    #
    # def disconnect(self) -> None:
    #     pass

    # INFO ------------------------------------------------------------------------------------------------------------
    # def dev__load_info(self) -> None:
    #     """
    #     GOAL
    #     ----
    #     load all important attrs in object.
    #     for further identification.

    #     WHERE
    #     -----
    #     useful in connect_validate
    #     """
    #     pass

    def dev__get_info(self) -> dict[str, Any]:
        """
        GOAL
        ----
        get already loaded data!
        """
        result = {
            "DEV_FOUND": self.DEV_FOUND,
            "INDEX": self.INDEX,
            "SKIP": self.SKIP,

            "NAME": self.NAME or self.__class__.__name__,
            "DESCRIPTION": self.DESCRIPTION or self.__class__.__name__,
            "SN": self.SN or "",
            "FW": self.FW or "",
            "MODEL": self.MODEL or "",

            "DUT_SN": self.DUT_SN or "",
            "DUT_FW": self.DUT_FW or "",
            "DUT_MODEL": self.DUT_MODEL or "",
        }
        return result


# =====================================================================================================================
class Base__DeviceUart_ETech(Base_Device, SerialClient_FirstFree_AnswerValid):
    """
    GOAL
    ----
    make a dev class for ETech UART devices

    NOTE
    ----
    Exc in SERIAL PORTs - connecting but not not working!
    ...VALUE_LINK=<function SerialClient.__getattr__.<locals>.<lambda> at 0x000001909BBF34C0>,ARGS__VALUE=('HV0',),value_last=Attempting to use a port that is not open,
    reason - some methods is over
    """
    LOG_ENABLE = True
    RAISE_CONNECT = False
    BAUDRATE = 115200
    EOL__SEND = b"\n"

    REWRITEIF_READNOANSWER = 0
    REWRITEIF_NOVALID = 0

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, index: int = None, **kwargs):
        """
        :param index: None is only for SINGLE! - TRY NOT TO USE IT!!!
        """
        if index is not None:
            self.INDEX = index
        super().__init__(**kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def PREFIX(self) -> str:
        return f"{self.NAME}:{self.INDEX+1:02d}:"

    def dev__load_info(self) -> None:
        if not self.SN:
            self.SN = self.write_read__last("get SN")
            self.FW = self.write_read__last("get FW")
            self.MODEL = self.write_read__last("get MODEL")

    # DETECT --------------------------------
    @property
    def DEV_FOUND(self) -> bool:
        return self.address_check__resolved()

    # ---------------------------------------
    def connect__validate(self) -> bool:
        result = self.address_check__resolved()  # fixme: is it really need here???

        if result:
            if self.NAME == "PTB":
                if not self.write_read__last_validate("get prsnt", "1"):
                    return False

            self.dev__load_info()

        return result

    # def address__validate(self) -> bool:  # NO NEED! only for manual!
    #     result = (
    #             self.write_read__last_validate("get name", self.NAME, prefix="*:")
    #             and
    #             self.write_read__last_validate("get addr", EqValid_NumParsedSingle_Success(self.INDEX+1), prefix="*:")
    #             # and
    #             # self.write_read__last_validate("get prsnt", "0")
    #     )
    #     if result:
    #         self.dev__load_info()
    #
    #     return result


# =====================================================================================================================
class Base_DeviceDummy(Base__DeviceUart_ETech):  # IMPORTANT! KEEP Serial FIRST Nesting!
    """
    GOAL
    ----
    use to make final Dummy
    """
    _ADDRESS = "DUMMY"

    @property
    def DEV_FOUND(self) -> bool:
        return True

    def address__validate(self) -> bool:
        return True

    def connect__validate(self) -> bool:
        return True

    def connect(self, *args, **kwargs) -> bool:
        return True


# =====================================================================================================================



# from typing import *
#
# from base_aux.aux_attr.m1_annot_attr1_aux import *
# from base_aux.base_nest_dunders.m1_init2_annots1_attrs_by_args_kwargs import *
# from base_aux.base_nest_dunders.m2_str2_attrs import *
# from base_aux.base_nest_dunders.m4_gsai_ic__annots import *
# from base_aux.base_nest_dunders.m6_eq1_attrs import *
# from base_aux.base_nest_dunders.m8_contains_attrs import *
# from base_aux.base_nest_dunders.m8_len_attrs import *
#
#
# # =====================================================================================================================
# @final
# class AttrKit_LockedKeys(       # DEPRECATED! - use DictIc_LockedKeys
#     NestInit_AnnotsAttr_ByKwargs,
#
#     NestGAI_AnnotAttrIC,
#
#     NestEq_AttrsNotHidden,
#     NestStR_AttrsNotHidden,
#     NestLen_AttrNotHidden,
#     NestContains_AttrIcNotHidden,
# ):
#     """
#     GOAL
#     ----
#     create attrs by init.
#     then try to update values.
#
#     SAME AS - 1=Base_AttrKit
#     ------------------------
#     BUT attrs create by stable names only on INIT by kwargs! online!
#
#     SPECIALLY CREATED FOR
#     ---------------------
#     Base_Indicator to use as inline params setup - like an objectDict! with exact attrSet!
#     """
#     def __init__(self, **kwargs: Any) -> None:
#         super().__init__(**kwargs)
#
#         AttrAux_AnnotsLast(self).annots__append_with_values(**kwargs)
#
#     def __call__(self, *args, **kwargs) -> Self | NoReturn:
#         """
#         GOAL
#         ----
#         set values only for existed keys
#         so it used as try to update def values!
#         """
#         AttrAux_AnnotsLast(self).sai__by_args(*args)
#
#         for name in kwargs:
#             if AttrAux_AnnotsLast(self).name_ic__get_original(name) is None:
#                 msg = f"{name=} not in {self=}"
#                 raise Exc__NotExistsNotFoundNotCreated(msg)
#         AttrAux_AnnotsLast(self).sai__by_kwargs(**kwargs)
#
#         return self
#
#
# # =====================================================================================================================
# if __name__ == '__main__':
#     pass
#
#
# # =====================================================================================================================



from typing import *

from base_aux.base_nest_dunders.m1_init1_source import *
from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_enums.m2_enum1_adj import *


# =====================================================================================================================
class StrIc(NestInit_Source):
    """
    NOTE
    ----
    DO NOT USE in simple dict as keys! - use exact DictIcKeys instead!

    GOAL
    ----
    cmp any object with IC meaning,
    i used this behaviour many times and finally decide to make an object

    SPECIALLY CREATED FOR
    ---------------------
    first creation for EnumEqValid to use for keys - but it is incorrect (not enough)!
    """
    SOURCE: str | Any | Callable[[], str | Any] = None
    RESTYLE: EnumAdj__TextCaseStyle = EnumAdj__TextCaseStyle.ORIGINAL   # REMAKE original source - todo: decide to deprecate?

    def _init_post(self) -> None:
        self.source_update()

    def source_update(self) -> None:
        """
        GOAL
        ----
        update/change source by expected style!

        WHY
        ---
        if you already have result on inition? - because smtms you can change source by adding some new data
        and after that you may be want toFix actual value
        """
        # resolve -----
        self.SOURCE = Lambda(self.SOURCE).resolve__exc()
        self.SOURCE = str(self.SOURCE)

        # restyle ------
        if self.RESTYLE == EnumAdj__TextCaseStyle.UPPER:
            self.SOURCE = self.SOURCE.upper()

        elif self.RESTYLE == EnumAdj__TextCaseStyle.LOWER:
            self.SOURCE = self.SOURCE.lower()

    def __str__(self) -> str:
        return str(self.SOURCE)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __eq__(self, other: Any) -> bool:
        return str(other).lower() == self.SOURCE.lower()

    def __hash__(self) -> int:
        return hash(self.SOURCE.lower())

    def __contains__(self, other: Any) -> bool:
        return str(other).lower() in self.SOURCE.lower()

    def __getitem__(self, item: int) -> Self | NoReturn:
        result = self.SOURCE.lower()[item]
        return self.__class__(result)

    def __len__(self) -> int:
        return len(str(self.SOURCE))

    def __iter__(self) -> Iterable[Self]:
        for item in self.SOURCE:
            yield self.__class__(item)


# =====================================================================================================================
class StrIcUpper(StrIc):
    RESTYLE = EnumAdj__TextCaseStyle.UPPER


# ---------------------------------------------------------------------------------------------------------------------
class StrIcLower(StrIc):
    RESTYLE = EnumAdj__TextCaseStyle.LOWER


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source_draft, _EXPECTED",
    argvalues=[
        (1, ["1", "1", "1"]),
        ("AaA", ["AaA", "AAA", "aaa"]),
        (Lambda("AaA"), ["AaA", "AAA", "aaa"]),
        (Lambda(1), ["1", "1", "1"]),
    ]
)
def test__1_str(source_draft, _EXPECTED):
    Lambda(lambda: str(StrIc(source_draft))).check_expected__assert(_EXPECTED[0])
    Lambda(lambda: str(StrIcUpper(source_draft))).check_expected__assert(_EXPECTED[1])
    Lambda(lambda: str(StrIcLower(source_draft))).check_expected__assert(_EXPECTED[2])


@pytest.mark.parametrize(
    argnames="source_draft, other_draft, _EXPECTED",
    argvalues=[
        (1, 1, True),
        (Lambda(1), 1, True),
        ("AaA", "aAa", True),
        (Lambda("AaA"), "aAa", True),
        (Lambda("AaA"), StrIc("aAa"), True),
    ]
)
def test__2_eq(source_draft, other_draft, _EXPECTED):
    Lambda(StrIc(source_draft) == other_draft).check_expected__assert(_EXPECTED)
    Lambda(StrIcUpper(source_draft) == other_draft).check_expected__assert(_EXPECTED)
    Lambda(StrIcLower(source_draft) == other_draft).check_expected__assert(_EXPECTED)


@pytest.mark.parametrize(
    argnames="source_draft, item, _EXPECTED",
    argvalues=[
        (1, 0, "1"),
        (1, 1, Exception),
        ("AaA", "aAa", Exception),
        (Lambda("AaA"), "aAa", Exception),
        (Lambda(1), 0, "1"),
        (Lambda(123), 0, "1"),
        (Lambda(123), 1, "2"),
        (Lambda("ABC"), 1, "b"),
    ]
)
def test__3_ga(source_draft, item, _EXPECTED):
    Lambda(lambda: StrIc(source_draft)[item]).check_expected__assert(_EXPECTED)
    Lambda(lambda: StrIcUpper(source_draft)[item]).check_expected__assert(_EXPECTED)
    Lambda(lambda: StrIcLower(source_draft)[item]).check_expected__assert(_EXPECTED)


@pytest.mark.parametrize(
    argnames="source_draft, other, _EXPECTED",
    argvalues=[
        (1, 0, False),
        (1, 1, True),
        ("AaA", "aAa", True),
        ("AaA", "Aa", True),
        (Lambda("AaA"), "Aa", True),
        (Lambda(1), 0, False),
        (Lambda(123), 0, False),
        (Lambda(123), 1, True),
    ]
)
def test__4_other_in(source_draft, other, _EXPECTED):
    Lambda(other in StrIc(source_draft)).check_expected__assert(_EXPECTED)
    Lambda(other in StrIcUpper(source_draft)).check_expected__assert(_EXPECTED)
    Lambda(other in StrIcLower(source_draft)).check_expected__assert(_EXPECTED)


@pytest.mark.parametrize(
    argnames="source_draft, items, _EXPECTED",
    argvalues=[
        (1, [0, 10], False),
        (1, [0, 1], True),
        ("AaA", ["aAa", "aA"], True),
        ("AaA", ["Aa", ], False),
        (Lambda("AaA"), ["Aa", ], False),
        (Lambda(1), [0, 10,],  False),
        (Lambda(123), [0, 10,], False),
        (Lambda(123), [0, 123,], True),
    ]
)
def test__5_in_other(source_draft, items, _EXPECTED):
    Lambda(StrIc(source_draft) in items).check_expected__assert(_EXPECTED)
    Lambda(StrIcUpper(source_draft) in items).check_expected__assert(_EXPECTED)
    Lambda(StrIcLower(source_draft) in items).check_expected__assert(_EXPECTED)


@pytest.mark.parametrize(
    argnames="source_1, source_2, _EXPECTED",
    argvalues=[
        (1, 0, 2),
        (1, 1, 1),
        ("AaA", "aAa", 1),
        ("AaA", "Aa", 2),
        (Lambda("AaA"), "Aa", 2),
        (Lambda(1), 0, 2),
        (Lambda(123), 0, 2),
    ]
)
def test__6_set(source_1, source_2, _EXPECTED):
    Lambda(len({StrIc(source_1), StrIcUpper(source_2)})).check_expected__assert(_EXPECTED)
    Lambda(len({StrIc(source_1), StrIcLower(source_2)})).check_expected__assert(_EXPECTED)
    Lambda(len({StrIcUpper(source_1), StrIcLower(source_2)})).check_expected__assert(_EXPECTED)


def test__10_gi_from_dict():
    assert {1: 1, StrIc("AAA"): 1, "2": 2}["aaa"] == 1
    try:
        assert {1: 1, StrIc("AAA"): 1, "2": 2}["AAA"] == 1
    except:
        print(f"THIS WILL newer work with gi! try use not a simple dict but DictIcKeys")
        return
    assert False


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



import pytest
import re

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_text.m0_patterns import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, fpoint, _EXP_int, _EXP_float, _EXP_both",
    argvalues=[
        # trash ----
        ("", None, (None, None), (None, None), (None, None)),
        ("hello", None, (None, None), (None, None), (None, None)),
        ("he.l,lo", None, (None, None), (None, None), (None, None)),

        ("11,22,33", None, (None, None), (None, None), (None, None)),
        ("11,22,33", ",", (None, None), (None, None), (None, None)),
        ("11.22.33", None, (None, None), (None, None), (None, None)),
        ("11.22.33", ",", (None, None), (None, None), (None, None)),

        # INT ------
        ("00100", None, ("00100", "00100"), (None, None), ("00100", "00100")),
        ("123", None, ("123", "123"), (None, None), ("123", "123")),
        ("a123b", None, (None, "123"), (None, None), (None, "123")),

        # FLOAT ----
        ("11,22", ".", (None, None), (None, None), (None, None)),
        ("11,22", ",", (None, None), ("11,22", "11,22"), ("11,22", "11,22")),
        ("11,22", None, (None, None), ("11,22", "11,22"), ("11,22", "11,22")),
        ("aa11,22bb", ",", (None, None), (None, "11,22"), (None, "11,22")),

        ("001.200", None, (None, None), ("001.200", "001.200"), ("001.200", "001.200")),
        ("11.22", None, (None, None), ("11.22", "11.22"), ("11.22", "11.22")),
        ("11.22", ",", (None, None), (None, None), (None, None)),

        # MINUS ----
        ("-123", None, ("-123", "-123"), (None, None), ("-123", "-123")),
        ("---123", None, (None, "-123"), (None, None), (None, "-123")),
        ("a-123a", None, (None, "-123"), (None, None), (None, "-123")),
        ("-a---123--a", None, (None, "-123"), (None, None), (None, "-123")),
        ("he-l,lo--11.22--=-asdf", None, (None, None), (None, "-11.22"), (None, "-11.22")),
    ]
)
def test___PatNumber(source, fpoint, _EXP_int, _EXP_float, _EXP_both):
    # INT -----
    match = re.fullmatch(Pat_NumberSingle(fpoint).INT_EXACT, source)
    Lambda(match and match[1]).check_expected__assert(_EXP_int[0])

    match = re.fullmatch(Pat_NumberSingle(fpoint).INT_COVERED, source)
    Lambda(match and match[1]).check_expected__assert(_EXP_int[1])

    # FLOAT -----
    match = re.fullmatch(Pat_NumberSingle(fpoint).FLOAT_EXACT, source)
    Lambda(match and match[1]).check_expected__assert(_EXP_float[0])

    match = re.fullmatch(Pat_NumberSingle(fpoint).FLOAT_COVERED, source)
    Lambda(match and match[1]).check_expected__assert(_EXP_float[1])

    # BOTH -----
    match = re.fullmatch(Pat_NumberSingle(fpoint).BOTH_EXACT, source)
    Lambda(match and match[1]).check_expected__assert(_EXP_both[0])

    match = re.fullmatch(Pat_NumberSingle(fpoint).BOTH_COVERED, source)
    Lambda(match and match[1]).check_expected__assert(_EXP_both[1])


# =====================================================================================================================



import time

from base_aux.testplans.tc import *
from base_aux.valid.m2_valid_derivatives import *
from base_aux.valid.m3_valid_chains import *
from .tc0_groups import *


# =====================================================================================================================
class TestCase(Base_TestCase):
    ASYNC = True
    DESCRIPTION = "PTB exist"

    @classmethod
    def startup__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        time.sleep(0.5)
        return True

    @classmethod
    def teardown__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        time.sleep(0.5)
        return True

    # RUN -------------------------------------------------------------------------------------------------------------
    def run__wrapped(self) -> TYPING__RESULT_W_EXC:
        result = ValidChains([
            time.sleep(1),
            Valid(
                value_link=self.DEV_COLUMN.DUT.address_check__resolved,
                name="DUT.address_check__resolved",
            ),
        ])
        return result


# =====================================================================================================================



from base_aux.aux_attr.m1_annot_attr1_aux import *
from base_aux.base_types.m1_type_aux import *
from base_aux.base_types.m0_static_types import *
from base_aux.base_types.m0_static_typing import *


# =====================================================================================================================
@final
class IterAux(NestInit_Source):
    """
    collect universal funcs which work with collections

    NOTE
    ----
    for access abilities passing with strings - resolve it by yourself

        assert self.victim("1/2", {1: 11, }) is None
        assert self.victim("1/2", {1: {2: 22}, }) == Explicit([1, 2, ])
        assert self.victim("1/2/1", {1: {2: [30, 31, 32]}, }) == Explicit([1, 2, 1])

        assert self.victim("hello", {"hello": [1]}) == Explicit(["hello", ])
        assert self.victim("hello/1", {"hello": [1]}) is None
        assert self.victim("hello/0", {"hello": [1]}) == Explicit(["hello", 0])

        assert self.victim("hello1/hello2", {"hello1": {"hello2": [1]}}) == Explicit(["hello1", "hello2"])
        assert self.victim("hello1/hello2/0", {"hello1": {"hello2": [1]}}) == Explicit(["hello1", "hello2", 0, ])
        assert self.victim("hello1/hello2/1", {"hello1": {"hello2": [1]}}) is None
    """
    SOURCE: TYPING.ITERABLE_ORDERED = dict
    # PATH: list[TYPING.ITERPATH_KEY]   # todo: get back!!! to work with source! or make new class!

    # def init_post(self):
    #     self.PATH = []

    # -----------------------------------------------------------------------------------------------------------------
    def item__get_original(self, item: Any | str | int, _raise: bool = None) -> Any | NoValue:
        """
        get FIRST original item from any collection by comparing str(expected).lower()==str(original).lower().

        # NOTE:
        # 1. NONE RESULT__VALUE - RESOLVED!!!
        # 2. SEVERAL VALUES - not used! by now it is just FIRST matched!
        #     several items? - it is not useful!!! returning first is most expected!
        #
        # USEFUL in case-insensitive systems (like terminals or serial devices) or object structured by prefix-names:
        # 1. get key in dict
        # 2. find attribute name in base_types
        #
        # :param item:
        # :return: actual item from collection
        #     None - if VALUE is unreachable/notFind
        """
        # if TypeAux(self.SOURCE).check__iterable_not_str():
        if isinstance(self.SOURCE, (list, tuple, dict, set)):
            values = self.SOURCE
        else:
            values = AttrAux_Existed(self.SOURCE).iter__names_filter__not_private()

        for value in values:
            try:
                if value == item or str(value).lower() == str(item).lower():
                    return value
            except:
                pass

        if _raise:
            raise AttributeError(f"{item=}")

        return NoValue

    # -----------------------------------------------------------------------------------------------------------------
    def item__check_exist(self, item: Any) -> bool:
        return self.item__get_original(item) is not NoValue

    # -----------------------------------------------------------------------------------------------------------------
    def keypath__get_original(self, *keypath: TYPING.ITERPATH_KEY) -> TYPING.ITERPATH | None | NoReturn:
        """
        NOTES:
        1. keypath used as address KEY for dicts and as INDEX for other listed data
        2. separator is only simple SLASH '/'!

        :param keypath:
        :return:
            None - if keypath is unreachable/incorrect
            tuple[Any] - reachable keypath which could be used to get VALUE from data by chain data[i1][i2][i3]
        """
        source = self.SOURCE
        if not keypath:
            return ()   # ROOT is OK!

        # work ----------------------------
        result = []
        for key_i in keypath:
            key_original = NoValue

            if isinstance(source, dict):
                key_original = IterAux(source).item__get_original(key_i)
                if key_original == NoValue:
                    return
                else:
                    source = source[key_original]

            elif isinstance(source, set):
                raise TypeError(f"{source=}")

            elif isinstance(source, (list, tuple)):
                try:
                    source = source[int(key_i)]
                    key_original = int(key_i)  # place last!
                except:
                    return

            else:
                key_original = AttrAux_Existed(source).name_ic__get_original(str(key_i))
                if key_original is None:
                    return
                else:
                    source = getattr(source, key_original)

            # -----------------------------
            result.append(key_original)

        return tuple(result)

    # -----------------------------------------------------------------------------------------------------------------
    def value__get(self, *keypath: TYPING.ITERPATH_KEY) -> Any | NoReturn:
        result = self.SOURCE
        keypath_orig = self.keypath__get_original(*keypath)
        if keypath_orig is None:
            print(f"{self.SOURCE=}/{keypath=}/{keypath_orig=}")
            raise Exc__NotExistsNotFoundNotCreated(f"{keypath=} in {self.SOURCE=}")
        for key_i in keypath_orig:
            try:
                result = result[key_i]
            except:
                result = AttrAux_Existed(result).gai_ic(key_i)     # raise

        return result

    def value__set(self, keypath: TYPING.ITERPATH, value: Any) -> bool:
        """
        GOAL
        ----
        INLINE WORK!
        """
        source = self.SOURCE

        # work ----------------------------
        keypath = self.keypath__get_original(*keypath)
        try:
            length = len(keypath)
            for pos, key_i in enumerate(keypath, start=1):
                if pos == length:
                    try:
                        source[key_i] = value
                    except:
                        AttrAux_Existed(source).sai_ic(key_i, value)     # raise
                    return True
                else:
                    source = IterAux(source).value__get(key_i)
        except:
            return False

    # -----------------------------------------------------------------------------------------------------------------
    def get_first_is_not(self, *variants: Any) -> Any | None:
        """
        GOAL
        ----
        from iterable get first is not None value!
        typically get value from params active/sub_default/default

        SPECIALY CREATED FOR
        --------------------
        Base_ReAttempts to get flags
        """
        for item in self.SOURCE:
            if not variants:
                return item

            if any([item is variant for variant in variants]):    # CMP ONLY BY IS_NOT_NONE!!! dont use '==' cause of __EQ__!
                continue
            else:
                return item

    def get_first_is_not_none(self) -> Any | None:
        return self.get_first_is_not(None)


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m1_init0_reinit2_lambdas_resolve import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="value, _EXPECTED",
    argvalues=[
        (1, True),
        ("hello", Exception),
    ]
)
def test__common__define(value, _EXPECTED):
    define_was_ok = True
    try:
        class Victim:
            ATTR0 = 0
            ATTR_INT = int(value)
            ATTR_STR = str(value)

    except Exception as exc:
        define_was_ok = exc

    Lambda(define_was_ok).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="value, _EXPECTED",
    argvalues=[
        (1, True),
        ("hello", Exception),
    ]
)
def test__special__define_and_init(value, _EXPECTED):
    # DEFINE ---------------------
    class Victim(NestInit_AttrsLambdaResolve):
        ATTR0 = 0
        ATTR_INT = Lambda(int, value)
        ATTR_STR = Lambda(str, value)

    assert True     # no exc above!

    # INIT -----------------------
    init_was_ok = True
    try:
        victim = Victim()
    except Exception as exc:
        init_was_ok = exc
    else:
        assert victim.ATTR0 == 0
        assert victim.ATTR_INT == value
        assert victim.ATTR_STR == str(value)

    Lambda(init_was_ok).check_expected__assert(_EXPECTED)


# =====================================================================================================================



from typing import *
import sys

import numpy as np
from numpy import dtype

from base_aux.base_values.m3_exceptions import *
from base_aux.aux_np_pd.m0_typing import *
from base_aux.base_nest_dunders.m1_init1_source import *


# =====================================================================================================================
# @final    # DONT USE FINAL!!! it used as base!
class NdArrayAux(NestInit_Source):
    SOURCE: TYPING__NP__FINAL

    # -----------------------------------------------------------------------------------------------------------------
    DEF__DTYPE_DICT: dict[str, str | dtype] = dict(  # template for making dtype
        # time='<i8',
        # open='<f8',
        # high='<f8',
        # low='<f8',
        # close='<f8',
        # tick_volume='<u8',
        # spread='<i4',
        # real_volume='<u8',
    )

    @property
    def DEF__DTYPE_ITEMS(self) -> list[tuple[str, str | dtype]]:
        """
        GOAL
        ----
        ref source final from draft if data struct is exact expected! like for TimeSeries!

        SPECIALLY CREATED FOR
        ---------------------
        NpTimeSeriesAux
        """
        return list(self.DEF__DTYPE_DICT.items())

    # -----------------------------------------------------------------------------------------------------------------
    def _init_post(self) -> None:
        self._init_post0__np_set_printoptions()
        self.unsure__ndarray_dtype()

    @staticmethod
    def _init_post0__np_set_printoptions():
        np.set_printoptions(threshold=sys.maxsize, linewidth=300)

    def unsure__ndarray_dtype(self) -> TYPING__NP__FINAL:
        """
        GOAL
        ----
        remake source if it not ndarray with names
        """
        if isinstance(self.SOURCE, (list, tuple)):
            if self.DEF__DTYPE_ITEMS:
                self.SOURCE = np.array(self.SOURCE, dtype=self.DEF__DTYPE_ITEMS)

        if not isinstance(self.SOURCE, np.ndarray):
            self.SOURCE = np.array(self.SOURCE)

        # TODO: make copy???
        return self.SOURCE  # return exact values just for tests!

    # -----------------------------------------------------------------------------------------------------------------
    def get_fields(self) -> dict[str, tuple[dtype, int]]:
        """
        GOAL
        ----
        just as help info!

        results
        -------
        DTYPE
            [('time', '<i8'), ('open', '<f8'), ('high', '<f8'), ('low', '<f8'), ('close', '<f8'), ('tick_volume', '<u8'), ('spread', '<i4'), ('real_volume', '<u8')]
            [
                ('time', '<i8'),
                ('open', '<f8'),
                ('high', '<f8'),
                ('low', '<f8'),
                ('close', '<f8'),
                ('tick_volume', '<u8'),
                ('spread', '<i4'),
                ('real_volume', '<u8')
            ]

        FIELDS
            ['time', 'open', 'high', 'low', 'close', 'tick_volume', 'spread', 'real_volume']

            {
                'time': (dtype('int64'), 0),
                'open': (dtype('float64'), 8),
                'high': (dtype('float64'), 16),
                'low': (dtype('float64'), 24),
                'close': (dtype('float64'), 32),
                'tick_volume': (dtype('uint64'), 40),
                'spread': (dtype('int32'), 48),
                'real_volume': (dtype('uint64'), 52)
            }
        """
        return self.SOURCE.dtype.fields


# =====================================================================================================================
class NdArray2dAux(NdArrayAux):
    # -----------------------------------------------------------------------------------------------------------------
    def d2_get_compact_str(
        self,
        values_translater: dict[Any, Any] = None,
        separate_rows_blocks: int = None,
        wrap: bool = None,
        use_rows_num: bool = None
    ) -> str:
        """
        :param values_translater: dictionary to change exact elements
        :param separate_rows_blocks: add blank line on step
        :param wrap: add additional strings before and after data
        :param use_rows_num: add row num in
        """
        values_translater = values_translater or {}
        count_rows = len(self.SOURCE)
        count_columns = len(self.SOURCE[0])
        row_pos = 0
        result: str = ""

        # tab_row_nums = (count_rows +1)//10 + 1
        # if (count_rows+1)%10 > 0:
        #     tab_row_nums += 1

        tab_row_nums = 4

        if wrap:
            if use_rows_num:
                result += " " * tab_row_nums
            result += "=" * count_columns + "\n"

        for row in self.SOURCE:
            row_pos += 1
            if separate_rows_blocks and row_pos > 1 and (row_pos - 1) % separate_rows_blocks == 0:
                result += f"\n"

            if use_rows_num:
                result += "{:{width}}".format(str(row_pos), width=tab_row_nums)
            for value in row:
                replaced = values_translater.get(value)
                if replaced is not None:
                    value = replaced
                result += f"{value}"

            if row_pos != count_rows:
                result += f"\n"

        if wrap:
            result += "\n"
            if use_rows_num:
                result += " " * tab_row_nums
            result += "=" * count_columns

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def split_groups(self, group_len: int) -> TYPING__NP__FINAL:
        """
        GOAL
        ----
        split array to arrays with exact elements count
        NOT INLINE! and dont do it!

        SPECIALLY CREATED FOR
        ---------------------
        FOR WHAT? --- delete????  it was trying to make windows and now not used????
        """
        # if self.SOURCE.ndim != 1:
        #     raise Exc__Incompatible_Data
        new_shape = []
        shape = self.SOURCE.shape
        a = np.arange(5)
        print(a)
        print(np.array_split(a, 2))
        print(a)
        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!

    # SHRINK ----------------------------------------------------------------------------------------------------------
    def shrink_simple(self, window_len: int) -> TYPING__NP__FINAL:
        """
        DIFFERENCE - from shrink
        ----------
        just drop other data! without calculations

        when important only one column in calculations!
        such as RSI/WMA typically use only close values from timeSeries!
        """
        return self.SOURCE[::window_len]

    def shrink(self, window_len: int) -> TYPING__NP__FINAL:
        """
        GOAL
        ----
        full remake TS to less TF then actual
        for example - you have 100history lines from tf=1m
        so you can get
            50lines for tf=2m
            10lines for tf=10m
        and it will be actual TS for TF! you dont need wait for finishing exact TF
        """
        if window_len == 1:
            return self.SOURCE
        elif window_len < 1:
            raise Exc__WrongUsage(f"{window_len=}")

        windows = self._windows_get(window_len)
        result = self._windows_shrink(windows)
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def _windows_get(self, window_len: int) -> TYPING__NP__FINAL:
        """
        GOAL
        ----
        separate ndarray into windows
        """
        bars_windows = np.lib.stride_tricks.sliding_window_view(x=self.SOURCE, window_shape=window_len)
        bars_windows_stepped = bars_windows[::window_len]
        return bars_windows_stepped

    @classmethod
    def _windows_shrink(cls, windows: np.ndarray) -> TYPING__NP__FINAL:
        result: Optional[np.ndarray] = None
        for window in windows:
            void_new = cls._window_shrink(window)
            try:
                result = np.concatenate([result, [void_new]])
            except Exception as exc:
                # if no elements
                # print(f"{exc!r}")
                result = np.array([void_new])
        return result

    @classmethod
    def _window_shrink(cls, window: np.ndarray) -> np.void | np.ndarray:   # NOTE: np.void - is acually! np.ndarray - just for IDE typeChecking!
        """
        GOAL
        ----
        modify several lines in exact one window into one line!

        SPECIALLY CREATED FOR
        ---------------------
        NpTimeSeriesAux.shrink
        """
        raise NotImplementedError()


# =====================================================================================================================



from base_aux.aux_eq.m2_eq_aux import *
from base_aux.base_types.m1_type_aux import *


# =====================================================================================================================
def _explore():
    class Cls:
        @classmethod
        def __eq__(cls, other):
            return True

    class Cls2(Cls):
        def __eq__(self, other):
            return True

    print(Cls == Cls2)  # False!
    print(Cls2 == Cls)  # False!


# =====================================================================================================================
class Nest_EqCls:
    """
    GOAL
    ----
    create ability to cmpEQ classes

    NOTE
    ----
    UNACCEPTABLE:
    1/ __eq__ is not working with classmethod!
    2/ meta is not acceptable!!!! in case of class is already Metaclassed like QThread!
        and in future it can become it!

    ACCEPTABLE
    3/ simply create method classmethod to cmp with any class
    4/ if you intend cmp only same classes - create propertyClassmethod

    BEST USAGE
    ----------
    1. create base class (with Nest_EqCls as parent)
    2. define _EQ_CLS__VALUE as classmethod to
    3. when need comparing classes use standard method _eq_cls__check(other) from any of object
    4. if you want cmp not comparring classes like Nest_EqCls-nested and not nested - redefine _eq_cls__check method manually  - but mayby it is not good idea!!!

    SPECIALLY CREATED FOR
    ---------------------
    TC cmp classes instead of MiddleGroup!
    """
    @classmethod
    def _eq_cls__check(cls, other: Any | type[Any]) -> bool:
        """
        show how to cmp clss
        """
        try:
            checkable = issubclass(other, Nest_EqCls)   # keep first!!!
        except:
            checkable = isinstance(other, Nest_EqCls)

        if checkable:
            return EqAux(cls._EQ_CLS__VALUE).check_doubleside__bool(other._EQ_CLS__VALUE)
        else:
            return False

    @classmethod
    @property
    def _EQ_CLS__VALUE(cls) -> Any:
        """
        GOAL
        ----
        REDEFINE TO USE AS CMP VALUE
        """
        return cls.__name__     # just as example and for zero comparing

    @classmethod
    def _eq_classes__check(cls, obj1: Any | type[Any], obj2: Any | type[Any]) -> bool:
        """
        eqCmp classes as function
        """
        cls1: Self = TypeAux(obj1).get__class()
        cls2: Self = TypeAux(obj2).get__class()

        if TypeAux(cls1).check__subclassed_or_isinst__from_cls_or_inst(Nest_EqCls):
            # cls1: Nest_EqCls
            return cls1._eq_cls__check(cls2)
        elif TypeAux(cls2).check__subclassed_or_isinst__from_cls_or_inst(Nest_EqCls):
            # cls1: Nest_EqCls
            return cls2._eq_cls__check(cls1)
        else:
            return cls1 == cls2 or cls2 == cls1


# =====================================================================================================================
def _best_usage():
    # with enums!
    from enum import Enum

    class ClsGroup(Enum):
        G1 = 1
        G2 = 2

    class Victim(Nest_EqCls):
        ATTR1 = 0
        @classmethod
        @property
        def _EQ_CLS__VALUE(cls) -> ClsGroup:
            """
            GOAL
            ----
            REDEFINE TO USE AS CMP VALUE
            """
            if cls.ATTR1 == 1:
                return ClsGroup.G1
            else:
                return ClsGroup.G2


# =====================================================================================================================
if __name__ == "__main__":
    _best_usage()


# =====================================================================================================================



from typing import *
import pytest

from base_aux.testplans.tp_manager import *


# =====================================================================================================================
class Test__Tp:
    @classmethod
    def setup_class(cls):
        pass

        class Victim(TpManager):
            pass

        cls.Victim = Victim

    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     passtest__tc.py
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.skip
    def test__1(self):
        victim = self.Victim()
        assert not victim.STAND.DEV_LINES
        assert not victim.INDEX

        assert not victim.result
        assert not victim.details
        assert not victim.exc

        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!


# =====================================================================================================================



from base_aux.aux_dict.m3_dict_ga import *
from base_aux.valid.m1_valid_base import *
from base_aux.pyqt.m0_signals import *

from base_aux.base_nest_dunders.m6_eq2_cls import *

from base_aux.testplans.tc_types import TYPING__RESULT_BASE, TYPING__RESULT_W_NORETURN, TYPING__RESULT_W_EXC
from base_aux.testplans.stand import *
from base_aux.loggers.m2_logger import *


# =====================================================================================================================
class _Base0_TestCase(Logger):
    """
    GOAL (SEPARATING)
    ----
    just to use in Signals before defining exact
    """
    pass


# =====================================================================================================================
class Enum_TcGroup_Base(NestEq_EnumAdj):
    NONE = None


# =====================================================================================================================
class Signals(Base_Signals):
    signal__tc_state_changed = pyqtSignal(_Base0_TestCase)


# =====================================================================================================================
class _Base1_TestCase(Nest_EqCls, _Base0_TestCase, QThread):
    LOG_ENABLE = False
    LOG_USE_FILE = False
    signals: Signals = Signals()  # FIXME: need signal ON BASE CLASS! need only one SlotConnection! need Singleton?

    # SETTINGS ------------------------------------
    NAME: str = ""      # set auto!
    DESCRIPTION: str = ""
    SKIP: Optional[bool] = None     # access only over CLASS attribute! not instance!!!
    skip_tc_dut: Optional[bool] = None
    ASYNC: Optional[bool] = True
    # STOP_IF_FALSE_RESULT: Optional[bool] = None     # NOT USED NOW! MAYBE NOT IMPORTANT!!!

    # AUXILIARY -----------------------------------
    STATE_ACTIVE__CLS: EnumAdj_ProcessStateActive = EnumAdj_ProcessStateActive.NONE

    result__startup_cls: TYPING__RESULT_BASE | EnumAdj_ProcessStateActive = None
    result__teardown_cls: TYPING__RESULT_BASE | EnumAdj_ProcessStateActive = None

    STAND: Base_Stand
    TCSi_LINE: TableLine = TableLine()

    # INSTANCE ------------------------------------
    INDEX: int
    SETTINGS: DictIcKeys_Ga = {}

    result__startup: TYPING__RESULT_W_EXC = None
    result__teardown: TYPING__RESULT_W_EXC = None

    _result: TYPING__RESULT_W_EXC = None
    timestamp_start: Optional[DateTimeAux] = None
    timestamp_stop: Optional[DateTimeAux] = None
    details: dict[str, Any]
    exc: Optional[Exception]

    # =================================================================================================================
    @property
    def DEV_COLUMN(self) -> TableColumn:
        return TableColumn(index=self.INDEX, lines=self.STAND.DEV_LINES)    # FIXME: use multiton!???

    # =================================================================================================================
    @classmethod
    @property
    def _EQ_CLS__VALUE(cls) -> Enum:
        """
        GOAL
        ----
        REDEFINE TO USE AS CMP VALUE
        """
        return Enum_TcGroup_Base.NONE

    # =================================================================================================================
    def __init__(self, index: int):
        self.INDEX = index
        self.clear()
        super().__init__()

    # =================================================================================================================
    @classmethod
    def clear__cls(cls):
        cls.STATE_ACTIVE__CLS = EnumAdj_ProcessStateActive.NONE
        cls.result__startup_cls = None
        cls.result__teardown_cls = None
        for tc in cls.TCSi_LINE:
            tc.clear()

    def clear(self) -> None:
        self.result__startup = None
        self.result__teardown = None
        self.result = None

        self.timestamp_start = None
        self.timestamp_stop = None

        self.details = {}
        self.exc = None

    # @classmethod
    # @property
    # def NAME(cls):
    #     return cls.__name__
    #     # return pathlib.Path(__file__).name    # work as last destination where property starts!

    # RESULT ----------------------------------------------------------------------------------------------------------
    @property
    def result(self) -> TYPING__RESULT_W_EXC:
        return self._result

    @result.setter
    def result(self, value: TYPING__RESULT_W_EXC) -> None:
        self._result = value
        self.signals.signal__tc_state_changed.emit(self)
        if isinstance(value, Exception):
            self.DEV_COLUMN.DUT.final_result: bool = False

        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!
        # FIXME: FINISH!!!

    # # ---------------------------------------------------------
    # @classmethod
    # @property
    # def result__startup_cls(cls) -> Optional[bool]:
    #     return cls.__result__cls_startup
    #
    # @classmethod
    # @result__startup_cls.setter
    # def result__startup_cls(cls, value: Optional[bool]) -> None:
    #     cls.__result__cls_startup = value
    #     # cls.signals.signal__tc_state_changed.emit(cls)

    # DETAILS ---------------------------------------------------------------------------------------------------------
    def details_update(self, details: dict[str, Any]) -> None:
        self.LOGGER.debug("")
        self.details.update(details)
        # self.signals.signal__tc_state_changed.emit(self)

    # =================================================================================================================
    @classmethod
    def run__cls(cls, cls_prev: type[Self] | None = None) -> None | bool:
        """run TC on batch duts(??? may be INDEXES???)
        preferred using in thread on upper level!

        :return:
            NONE - if SKIP for any reason
            True - need continue TP
            False - cant continue! need stop TP
        """
        # if not cls.STAND.DEV_LINES.DUT:
        #     return

        print(f"run__cls=START={cls.NAME=}/{cls.DESCRIPTION=}{'=' * 50}")

        # SKIP ---------------------------------------------------
        # if cls.SKIP:
        #     print(f"run__cls=SKIP={cls.NAME=}={'=' * 50}")
        #     return

        cls.clear__cls()
        cls.STATE_ACTIVE__CLS = EnumAdj_ProcessStateActive.STARTED

        # FIXME: teardown not call!!!

        # TERDOWN PREV ----------------------------------------
        if cls_prev and not Nest_EqCls._eq_classes__check(cls, cls_prev):
            cls_prev.result__teardown_cls = cls_prev.teardown__cls()
            if cls_prev.result__startup_cls and cls_prev.result__teardown_cls is False:
                return False

        # STARTUP ----------------------------------------
        if cls_prev is not None and Nest_EqCls._eq_classes__check(cls, cls_prev):
            cls.result__startup_cls = cls_prev.result__startup_cls
        elif not Nest_EqCls._eq_classes__check(cls, cls_prev):
            cls.result__startup_cls = cls.startup__cls()

        # WORK ---------------------------------------------------
        if cls.result__startup_cls:
            # BATCH --------------------------
            for tc_inst in cls.TCSi_LINE:
                if tc_inst.skip_tc_dut:
                    continue

                print(f"run__cls=tc_inst.start({tc_inst.INDEX=})")
                tc_inst.start()
                if not cls.ASYNC:
                    print(f"run__cls=tc_inst.wait({tc_inst.INDEX=})inONEbyONE")
                    tc_inst.wait()

            # WAIT --------------------------
            if cls.ASYNC:
                for tc_inst in cls.TCSi_LINE:
                    print(f"run__cls=tc_inst.wait({tc_inst.INDEX=})inPARALLEL")
                    tc_inst.wait()

        # FINISH -------------------------------------------------
        cls.STATE_ACTIVE__CLS = EnumAdj_ProcessStateActive.FINISHED
        print(f"[TC]FINISH={cls.NAME=}={'=' * 50}")
        return True

    def run(self) -> None:
        """
        GOAL
        ----
        start execution INSTANCE (in thread)
        """
        self.LOGGER.debug("run")

        # PREPARE --------
        self.clear()
        self.timestamp_start = DateTimeAux()
        if (
                not hasattr(self.DEV_COLUMN, "DUT")
                or
                self.DEV_COLUMN.DUT.SKIP
                or
                not self.DEV_COLUMN.DUT.DEV_FOUND
                or
                not self.DEV_COLUMN.DUT.connect()
        ):
            return

        # WORK --------
        self.LOGGER.debug("run-startup")
        if self.startup():
            try:
                self.LOGGER.debug("run-run_wrapped START")
                self.result = self.run__wrapped()
                if isinstance(self.result, Valid):
                    self.result.run__if_not_finished()

                self.LOGGER.debug(f"run-run_wrapped FINISHED WITH {self.result=}")
            except Exception as exc:
                self.result = False
                self.exc = exc
        self.LOGGER.debug("run-teardown")
        self.teardown()

    # =================================================================================================================
    @classmethod
    def startup__cls(cls) -> TYPING__RESULT_W_EXC:
        """before batch work
        """
        print(f"startup__cls")
        cls.result__startup_cls = EnumAdj_ProcessStateActive.STARTED
        # cls.clear__cls()

        result = cls.startup__cls__wrapped
        result = Lambda(result).resolve__exc()
        if isinstance(result, Valid):
            result.run__if_not_finished()
        print(f"{cls.result__startup_cls=}")
        cls.result__startup_cls = result
        return result

    def startup(self) -> TYPING__RESULT_W_EXC:
        self.LOGGER.debug("")

        result = self.startup__wrapped
        result = Lambda(result).resolve__exc()
        if isinstance(result, Valid):
            result.run__if_not_finished()
        self.result__startup = result
        return result

    def teardown(self) -> TYPING__RESULT_W_EXC:
        self.LOGGER.debug("")
        self.timestamp_stop = DateTimeAux()

        result = self.teardown__wrapped
        result = Lambda(result).resolve__exc()
        if isinstance(result, Valid):
            result.run__if_not_finished()

        self.result__teardown = result
        return result

    @classmethod
    def teardown__cls(cls) -> TYPING__RESULT_W_EXC:
        print(f"run__cls=teardown__cls")

        if cls.STATE_ACTIVE__CLS == EnumAdj_ProcessStateActive.STARTED or cls.result__teardown_cls is None:
            print(f"run__cls=teardown__cls=1")
            cls.result__teardown_cls = Lambda(cls.teardown__cls__wrapped).resolve__exc()
            if isinstance(cls.result__teardown_cls, Valid):
                cls.result__teardown_cls.run__if_not_finished()

            if not bool(cls.result__teardown_cls):
                print(f"[FAIL]{cls.result__teardown_cls=}//{cls.NAME}")

        else:
            print(f"run__cls=teardown__cls=2")
            pass

        print(f"run__cls=teardown__cls=3")
        cls.STATE_ACTIVE__CLS = EnumAdj_ProcessStateActive.FINISHED
        return cls.result__teardown_cls

    # =================================================================================================================
    @classmethod
    def terminate__cls(cls) -> None:
        for tc_inst in cls.TCSi_LINE:
            try:
                if tc_inst.isRunning() and not tc_inst.isFinished():
                    tc_inst.terminate()
            except:
                pass

        cls.teardown__cls()

    # -----------------------------------------------------------------------------------------------------------------
    def terminate(self) -> None:
        self.LOGGER.debug("")
        super().terminate()
        self.teardown()     # TODO: check order!!! place upper!

    # REDEFINE ========================================================================================================
    pass
    pass
    pass
    pass
    pass
    pass

    @classmethod
    def startup__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        return True

    def startup__wrapped(self) -> TYPING__RESULT_W_NORETURN:
        return True

    def run__wrapped(self) -> TYPING__RESULT_W_NORETURN:
        return True

    def teardown__wrapped(self) -> TYPING__RESULT_W_NORETURN:
        return True

    @classmethod
    def teardown__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        return True


# =====================================================================================================================
class _Info(_Base1_TestCase):
    """
    separated class for gen results/info by models!
    """
    INFO_STR__ADD_ATTRS: Iterable[str] = []

    @classmethod
    def tcc__get_info(cls) -> dict[str, Any]:
        """
        GOAL
        ----
        get info about TcCls
        """
        result = {
            "TC_NAME": cls.NAME,
            "TC_DESCRIPTION": cls.DESCRIPTION,
            "TC_ASYNC": cls.ASYNC,
            "TC_SKIP": cls.SKIP,
        }
        return result

    # =================================================================================================================
    def tci__get_results__pretty(self) -> str:
        # FIXME: GET FROM INFO_GET????
        result = ""

        result += f"INDEX={self.INDEX}\n"
        result += f"DUT_SN={self.DEV_COLUMN.DUT.SN}\n"
        result += f"DUT_ADDRESS={self.DEV_COLUMN.DUT.ADDRESS}\n"
        result += f"tc_skip_dut={self.skip_tc_dut}\n"

        result += f"TC_NAME={self.NAME}\n"
        result += f"TC_GROUP={self._EQ_CLS__VALUE}\n"
        result += f"TC_DESCRIPTION={self.DESCRIPTION}\n"
        result += f"TC_ASYNC={self.ASYNC}\n"
        result += f"TC_SKIP={self.SKIP}\n"

        result += f"SETTINGS=====================\n"
        if self.SETTINGS:
            for name, value in self.SETTINGS.items():
                result += f"{name}:{value}\n"

        result += f"INFO_STR__ADD_ATTRS===========\n"
        if self.INFO_STR__ADD_ATTRS:
            for name in self.INFO_STR__ADD_ATTRS:
                if hasattr(self, name):
                    result += f"{name}:{getattr(self, name)}\n"

        result += f"PROGRESS=====================\n"
        result += f"STATE_ACTIVE__CLS={self.__class__.STATE_ACTIVE__CLS}\n"
        result += f"timestamp_start={self.timestamp_start}\n"
        result += f"timestamp_stop={self.timestamp_stop}\n"
        result += f"exc={self.exc}\n"

        result += "-"*60 + "\n"
        result += f"result__startup={self.result__startup}\n"
        result += f"result={self.result}\n"
        result += f"result__teardown={self.result__teardown}\n"
        result += "-"*60 + "\n"

        result += f"DETAILS=====================\n"
        for name, value in self.details.items():
            result += f"{name}={value}\n"
        return result

    # =================================================================================================================
    @classmethod
    def tcsi__get_results(cls) -> dict[int, dict[str, Any]]:
        results = {}
        for tc_inst in cls.TCSi_LINE:
            results.update({tc_inst.INDEX: tc_inst.tci__get_result()})
        return results

    # -----------------------------------------------------------------------------------------------------------------
    def tci__get_result(self, add_info_dut: bool = True, add_info_tc: bool = True) -> dict[str, Any]:
        self.LOGGER.debug("")

        info_dut = {}
        try:
            if add_info_dut:
                info_dut = self.DEV_COLUMN.DUT.dev__get_info()
        except:
            pass

        info_tc = {}
        if add_info_tc:
            info_tc = self.tcc__get_info()

        result = {
            **info_tc,
            **info_dut,

            # RESULTS
            "timestamp_start": self.timestamp_start and str(self.timestamp_start),
            "tc_active": self.isRunning(),
            "tc_result_startup": bool(self.result__startup),
            "tc_result": None if self.result is None else bool(self.result),
            "tc_details": self.details,
            "result__teardown": bool(self.result__teardown),
            "timestamp_stop": self.timestamp_stop and str(self.timestamp_stop),
            "log": self.tci__get_results__pretty().replace("\"", "").replace("\'", ""),
        }
        return result


# =====================================================================================================================
class Base_TestCase(_Info):
    """
    """


# =====================================================================================================================
